"""Instrument pylibmc to report Memcached queries.

``patch_all`` will automatically patch your pylibmc client to make it work.
::

    # Be sure to import pylibmc and not pylibmc.Client directly,
    # otherwise you won't have access to the patched version
    from ddtrace import Pin, patch
    import pylibmc

    # If not patched yet, you can patch pylibmc specifically
    patch(pylibmc=True)

    # One client instrumented with default configuration
    client = pylibmc.Client(["localhost:11211"]
    client.set("key1", "value1")

    # Use a pin to specify metadata related to this client
    Pin.override(client, service="memcached-sessions")
"""

from ...internal.utils.importlib import require_modules


required_modules = ["pylibmc"]

with require_modules(required_modules) as missing_modules:
    if not missing_modules:
        from .client import TracedClient
        from .patch import patch

        __all__ = ["TracedClient", "patch"]



import os
import re
from typing import List
from typing import Optional
from typing import Tuple


def path_to_regex(pattern):
    # type: (str) -> re.Pattern
    """
    source https://github.com/sbdchd/codeowners/blob/c95e13d384ac09cfa1c23be1a8601987f41968ea/codeowners/__init__.py

    Copyright (c) 2019-2020 Steve Dignam

    Permission is hereby granted, free of charge, to any person obtaining
    a copy of this software and associated documentation files (the
    "Software"), to deal in the Software without restriction, including
    without limitation the rights to use, copy, modify, merge, publish,
    distribute, sublicense, and/or sell copies of the Software, and to
    permit persons to whom the Software is furnished to do so, subject to
    the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
    LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
    OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
    WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

    ported from https://github.com/hmarr/codeowners/blob/d0452091447bd2a29ee508eebc5a79874fb5d4ff/match.go#L33

    MIT License

    Copyright (c) 2020 Harry Marr
    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:
    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.
    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
    """
    regex = ""

    slash_pos = pattern.find("/")
    anchored = slash_pos > -1 and slash_pos != len(pattern) - 1

    regex += r"\A" if anchored else r"(?:\A|/)"

    matches_dir = pattern[-1] == "/"
    pattern_trimmed = pattern.strip("/")

    in_char_class = False
    escaped = False

    iterator = enumerate(pattern_trimmed)
    for i, ch in iterator:

        if escaped:
            regex += re.escape(ch)
            escaped = False
            continue

        if ch == "\\":
            escaped = True
        elif ch == "*":
            if i + 1 < len(pattern_trimmed) and pattern_trimmed[i + 1] == "*":
                left_anchored = i == 0
                leading_slash = i > 0 and pattern_trimmed[i - 1] == "/"
                right_anchored = i + 2 == len(pattern_trimmed)
                trailing_slash = i + 2 < len(pattern_trimmed) and pattern_trimmed[i + 2] == "/"

                if (left_anchored or leading_slash) and (right_anchored or trailing_slash):
                    regex += ".*"

                    next(iterator, None)
                    next(iterator, None)
                    continue
            regex += "[^/]*"
        elif ch == "?":
            regex += "[^/]"
        elif ch == "[":
            in_char_class = True
            regex += ch
        elif ch == "]":
            if in_char_class:
                regex += ch
                in_char_class = False
            else:
                regex += re.escape(ch)
        else:
            regex += re.escape(ch)

    if in_char_class:
        raise ValueError("unterminated character class in pattern {pattern}".format(pattern=pattern))

    regex += "/" if matches_dir else r"(?:\Z|/)"
    return re.compile(regex)


class Codeowners(object):
    """Provide interface to parse CODEOWNERS file and match a given path against it."""

    KNOWN_LOCATIONS = (
        "CODEOWNERS",
        ".github/CODEOWNERS",
        "docs/CODEOWNERS",
        ".gitlab/CODEOWNERS",
    )

    def __init__(self, path=None, cwd=None):
        # type: (Optional[str], Optional[str]) -> None
        """Initialize Codeowners object.

        :param path: path to CODEOWNERS file otherwise try to use any from known locations
        """
        path = path or self.location(cwd)
        if path is not None:
            self.path = path  # type: str
        self.patterns = []  # type: List[Tuple[re.Pattern, List[str]]]
        self.parse()

    def location(self, cwd=None):
        # type: (Optional[str]) -> Optional[str]
        """Return the location of the CODEOWNERS file."""
        cwd = cwd or os.getcwd()
        for location in self.KNOWN_LOCATIONS:
            path = os.path.join(cwd, location)
            if os.path.isfile(path):
                return path
        raise ValueError("CODEOWNERS file not found")

    def parse(self):
        # type: () -> None
        """Parse CODEOWNERS file and store the lines and regexes."""
        with open(self.path) as f:
            patterns = []
            for line in f.readlines():
                line = line.strip()
                if line == "":
                    continue
                # Lines starting with '#' are comments.
                if line.startswith("#"):
                    continue
                if line.startswith("[") and line.endswith("]"):
                    # found a code owners section
                    continue
                if line.startswith("^[") and line.endswith("]"):
                    # found an optional code owners section
                    continue

                elements = line.split()
                if len(elements) < 2:
                    continue

                path = elements[0]
                if path is None:
                    continue

                try:
                    pattern = path_to_regex(path)
                except (ValueError, IndexError):
                    continue

                owners = [owner for owner in elements[1:] if owner]

                if not owners:
                    continue
                patterns.append((pattern, owners))
            # Order is important. The last matching pattern has the most precedence.
            patterns.reverse()
            self.patterns = patterns

    def of(self, path):
        # type: (str) -> List[str]
        """Return code owners for a given path.

        :param path: path to check
        :return: list of file code owners identified by the given path
        """
        for pattern, owners in self.patterns:
            if pattern.search(path):
                return owners
        raise KeyError("no code owners found for {path}".format(path=path))



def __specialstuff__(arg):
    return arg


def modulestuff(snafu):
    return snafu


def decorator(f):
    def identity(*args, **kwargs):
        return f(*args, **kwargs)

    return identity


def decoratorwitharg(arg):
    def decorator(f):
        def foo(arg, *args, **kwargs):
            return arg, f(*args, **kwargs)

        return foo

    return decorator


class Stuff(object):
    @staticmethod
    def staticstuff(foo):
        return foo

    @classmethod
    def classstuff(cls):
        return cls

    def instancestuff(self, bar=None):
        return bar

    @property
    def propertystuff(self):
        return -1

    @propertystuff.setter
    def propertystuff(self, value):
        return value

    @decorator
    def decoratedstuff(self):
        return self

    @decorator
    @decorator
    def doublydecoratedstuff(self):
        return None

    @decoratorwitharg(42)
    def decoratedwithargsstuff(self):
        return self

    def nestedstuff(self):
        def localstuff(arg):
            def localerstuff(arg):
                return arg

            return localerstuff(arg)

        return localstuff(self)

    def pointlessstuff(self):
        # TODO: This needs to include the line numbers of its code objects.
        # However, something like this is pretty useless.
        def pointlesslocal(foo):
            return foo or self or pointlesslocal

    def __mangledstuff(self):
        return self

    def generatorstuff(self, n):
        yield "Ready"
        # yield from range(n)
        yield "Done"
        return


class MoreStuff(Stuff):
    def __init__(self):
        self.foo = "foo"

    def hellostuff(self):
        return self.foo


def excstuff():
    try:
        raise Exception("Hello", "world!", 42)
    except Exception:
        pass


alias = modulestuff


class AliasStuff(object):
    def foo(self):
        pass

    bar = foo


def throwexcstuff():
    raise Exception("Hello", "world!", 42)


# TODO: We don't support lambdas for the same reasons we don't support local
# functions.
def lambdastuff():
    return (lambda x: x << 1)(21)


class PropertyStuff(object):
    import operator

    def __init__(self):
        self._foo = "foo"

    foo = property(operator.attrgetter("_foo"))


from time import sleep  # noqa


def durationstuff(ns):

    sleep(ns / 1e9)



from collections import defaultdict
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from ddtrace.debugging._probe.model import FunctionProbe
from ddtrace.debugging._probe.model import LineProbe
from ddtrace.debugging._probe.model import Probe
from ddtrace.debugging._probe.status import ProbeStatusLogger
from ddtrace.internal import forksafe
from ddtrace.internal.compat import ExcInfoType
from ddtrace.internal.logger import get_logger


logger = get_logger(__name__)


class ProbeRegistryEntry(object):

    __slots__ = ("probe", "installed", "exc_info", "message")

    def __init__(self, probe):
        # type: (Probe) -> None
        self.probe = probe
        self.installed = False
        self.exc_info = None  # type: Optional[ExcInfoType]
        self.message = None  # type: Optional[str]

    def activate(self):
        # type: () -> None
        return self.probe.activate()

    def deactivate(self):
        # type: () -> None
        return self.probe.deactivate()

    def set_installed(self):
        # type: () -> None
        self.installed = True

    def set_exc_info(self, exc_info):
        # type: (ExcInfoType) -> None
        self.exc_info = exc_info

    def set_message(self, message):
        # type: (str) -> None
        self.message = message


def _get_probe_location(probe):
    # type: (Probe) -> Optional[str]
    if isinstance(probe, LineProbe):
        return probe.source_file
    elif isinstance(probe, FunctionProbe):
        return probe.module
    else:
        raise ValueError("Unsupported probe type: {}".format(type(probe)))


class ProbeRegistry(dict):
    """Keep track of all the registered probes.

    New probes are also registered as pending, on a location basis, until they
    are processed (e.g. installed, generally by some import hook). Pending
    probes can be retrieved with the ``get_pending`` method.
    """

    def __init__(self, status_logger, *args, **kwargs):
        # type: (ProbeStatusLogger, Any, Any) -> None
        """Initialize the probe registry."""
        super(ProbeRegistry, self).__init__(*args, **kwargs)
        self.logger = status_logger

        # Used to keep track of probes pending installation
        self._pending = defaultdict(list)  # type: Dict[str, List[Probe]]

        self._lock = forksafe.RLock()

    def register(self, *probes):
        # type: (Probe) -> None
        """Register a probe."""
        with self._lock:
            for probe in probes:
                if probe in self:
                    # Already registered.
                    continue

                self[probe.probe_id] = ProbeRegistryEntry(probe)

                location = _get_probe_location(probe)
                if location is None:
                    self.set_error(probe, "Unable to resolve location information for probe {}".format(probe.probe_id))
                    continue

                self._pending[location].append(probe)

                self.logger.received(probe)

    def set_installed(self, probe):
        # type: (Probe) -> None
        """Set the installed flag for a probe."""
        with self._lock:
            self[probe.probe_id].set_installed()

            # No longer pending
            self._remove_pending(probe)

            self.logger.installed(probe)

    def set_exc_info(self, probe, exc_info):
        # type: (Probe, ExcInfoType) -> None
        """Set the installed flag for a probe."""
        with self._lock:
            self[probe.probe_id].set_exc_info(exc_info)
            self.logger.error(probe, exc_info=exc_info)

    def set_error(self, probe, message):
        # type: (Probe, str) -> None
        """Set the error message for a probe."""
        with self._lock:
            self[probe.probe_id].set_message(message)
            self.logger.error(probe, message)

    def log_probes_status(self):
        # type: () -> None
        """Log the status of all the probes using the status logger."""
        with self._lock:
            for entry in self.values():
                if entry.installed:
                    self.logger.installed(entry.probe)
                elif entry.exc_info:
                    self.logger.error(entry.probe, exc_info=entry.exc_info)
                elif entry.message:
                    self.logger.error(entry.probe, message=entry.message)
                else:
                    self.logger.received(entry.probe)

    def _remove_pending(self, probe):
        # type: (Probe) -> None
        location = _get_probe_location(probe)

        # Pending probes must have valid location information
        assert location is not None, probe

        pending_probes = self._pending[location]
        try:
            # DEV: Note that this is O(n), which is fine with a conservative
            # number of probes.
            pending_probes.remove(probe)
        except ValueError:
            # The probe wasn't pending
            pass
        if not pending_probes:
            del self._pending[location]

    def has_probes(self, location):
        # type: (str) -> bool
        for entry in self.values():
            if _get_probe_location(entry.probe) == location:
                return True
        return False

    def unregister(self, *probes):
        # type: (Probe) -> List[Probe]
        """Unregister a collection of probes.

        This also ensures that any pending probes are removed if they haven't
        been processed yet.
        """
        unregistered_probes = []
        with self._lock:
            for probe in probes:
                try:
                    entry = self.pop(probe.probe_id)
                except KeyError:
                    # We don't seem to have the probe
                    logger.warning("Tried to unregister unregistered probe %s", probe.probe_id)
                else:
                    probe = entry.probe
                    self._remove_pending(probe)
                    unregistered_probes.append(probe)
        return unregistered_probes

    def get_pending(self, location):
        # type: (str) -> List[Probe]
        """Get the currently pending probes by location."""
        return self._pending[location]

    def __contains__(self, probe):
        # type: (object) -> bool
        """Check if a probe is in the registry."""
        assert isinstance(probe, Probe)

        with self._lock:
            return super(ProbeRegistry, self).__contains__(probe.probe_id)



import asyncio

import aioredis
import pytest

from ddtrace import Pin
from ddtrace import tracer
from ddtrace.contrib.aioredis.patch import aioredis_version
from ddtrace.contrib.aioredis.patch import patch
from ddtrace.contrib.aioredis.patch import unpatch
from ddtrace.vendor.wrapt import ObjectProxy
from tests.utils import override_config

from ..config import REDIS_CONFIG


@pytest.mark.asyncio
@pytest.fixture
async def redis_client():
    r = await get_redis_instance(max_connections=10)  # default values
    yield r


@pytest.mark.asyncio
@pytest.fixture
async def single_pool_redis_client():
    r = await get_redis_instance(max_connections=1)
    yield r


def get_redis_instance(max_connections: int):
    if aioredis_version >= (2, 0):
        return aioredis.from_url("redis://127.0.0.1:%s" % REDIS_CONFIG["port"], max_connections=max_connections)
    return aioredis.create_redis_pool(("127.0.0.1", REDIS_CONFIG["port"]), maxsize=max_connections)


@pytest.mark.asyncio
@pytest.fixture(autouse=True)
async def traced_aioredis(redis_client):
    await redis_client.flushall()

    patch()
    try:
        yield
    finally:
        unpatch()
    await redis_client.flushall()


def test_patching():
    """
    When patching aioredis library
        We wrap the correct methods
    When unpatching aioredis library
        We unwrap the correct methods
    """
    if aioredis_version >= (2, 0):
        assert isinstance(aioredis.client.Redis.execute_command, ObjectProxy)
        assert isinstance(aioredis.client.Redis.pipeline, ObjectProxy)
        assert isinstance(aioredis.client.Pipeline.pipeline, ObjectProxy)
        unpatch()
        assert not isinstance(aioredis.client.Redis.execute_command, ObjectProxy)
        assert not isinstance(aioredis.client.Redis.pipeline, ObjectProxy)
        assert not isinstance(aioredis.client.Pipeline.pipeline, ObjectProxy)
    else:
        assert isinstance(aioredis.Redis.execute, ObjectProxy)
        assert isinstance(aioredis.commands.Redis.execute, ObjectProxy)
        unpatch()
        assert not isinstance(aioredis.Redis.execute, ObjectProxy)
        assert not isinstance(aioredis.commands.Redis.execute, ObjectProxy)


@pytest.mark.asyncio
@pytest.mark.snapshot
async def test_basic_request(redis_client):
    val = await redis_client.get("cheese")
    assert val is None


@pytest.mark.asyncio
@pytest.mark.snapshot
async def test_unicode_request(redis_client):
    val = await redis_client.get(u"😐")
    assert val is None


@pytest.mark.asyncio
@pytest.mark.snapshot
async def test_decoding_non_utf8_args(redis_client):
    await redis_client.set(b"\x80foo", b"\x80abc")
    val = await redis_client.get(b"\x80foo")
    assert val == b"\x80abc"


@pytest.mark.asyncio
@pytest.mark.snapshot(variants={"": aioredis_version >= (2, 0), "13": aioredis_version < (2, 0)})
async def test_decoding_non_utf8_pipeline_args(redis_client):
    if aioredis_version >= (2, 0):
        p = await redis_client.pipeline(transaction=False)
        await p.set(b"\x80blah", "boo")
        await p.set("foo", b"\x80abc")
        await p.get(b"\x80blah")
        await p.get("foo")
    else:
        p = redis_client.pipeline()
        p.set(b"\x80blah", "boo")
        p.set("foo", b"\x80abc")
        p.get(b"\x80blah")
        p.get("foo")

    response_list = await p.execute()
    assert response_list[0] is True  # response from redis.set is OK if successfully pushed
    assert response_list[1] is True
    assert response_list[2].decode() == "boo"
    assert response_list[3] == b"\x80abc"


@pytest.mark.skipif(aioredis_version > (2, 0), reason="only affects aioredis < 2.0")
@pytest.mark.asyncio
@pytest.mark.snapshot
async def test_closed_connection_pool(single_pool_redis_client):
    """
    Make sure it doesn't raise error when no free connections are available.
    After aioredis 2.0 it raises Too many connections error when it happens.
    """

    async def execute_task():
        """Execute call in the background as a task."""
        return single_pool_redis_client.get("cheese")

    # start running the task after blocking the pool
    with (await single_pool_redis_client):
        task = [asyncio.ensure_future(execute_task())]
    # Pool is released, make sure we wait for the task to finish
    await asyncio.gather(*task, return_exceptions=True)


@pytest.mark.asyncio
@pytest.mark.snapshot
async def test_long_command(redis_client):
    length = 1000
    val_list = await redis_client.mget(*range(length))
    assert len(val_list) == length
    for val in val_list:
        assert val is None


@pytest.mark.asyncio
@pytest.mark.snapshot
async def test_override_service_name(redis_client):
    with override_config("aioredis", dict(service_name="myaioredis")):
        val = await redis_client.get("cheese")
        assert val is None
        await redis_client.set("cheese", "my-cheese")
        val = await redis_client.get("cheese")
        if isinstance(val, bytes):
            val = val.decode()
        assert val == "my-cheese"


@pytest.mark.asyncio
@pytest.mark.snapshot
async def test_pin(redis_client):
    Pin.override(redis_client, service="my-aioredis")
    val = await redis_client.get("cheese")
    assert val is None


@pytest.mark.asyncio
@pytest.mark.snapshot(variants={"": aioredis_version >= (2, 0), "13": aioredis_version < (2, 0)})
async def test_pipeline_traced(redis_client):
    if aioredis_version >= (2, 0):
        p = await redis_client.pipeline(transaction=False)
        await p.set("blah", "boo")
        await p.set("foo", "bar")
        await p.get("blah")
        await p.get("foo")
    else:
        p = redis_client.pipeline()
        p.set("blah", "boo")
        p.set("foo", "bar")
        p.get("blah")
        p.get("foo")

    response_list = await p.execute()
    assert response_list[0] is True  # response from redis.set is OK if successfully pushed
    assert response_list[1] is True
    assert (
        response_list[2].decode() == "boo"
    )  # response from hset is 'Integer reply: The number of fields that were added.'
    assert response_list[3].decode() == "bar"


@pytest.mark.skipif(aioredis_version < (2, 0), reason="only supported in aioredis >= 2.0")
@pytest.mark.asyncio
@pytest.mark.snapshot
async def test_pipeline_traced_context_manager_transaction(redis_client):
    """
    Regression test for: https://github.com/DataDog/dd-trace-py/issues/3106

    https://aioredis.readthedocs.io/en/latest/migration/#pipelines-and-transactions-multiexec

    Example::

        async def main():
            redis = await aioredis.from_url("redis://localhost")
            async with redis.pipeline(transaction=True) as pipe:
                ok1, ok2 = await (pipe.set("key1", "value1").set("key2", "value2").execute())
            assert ok1
            assert ok2
    """

    async with redis_client.pipeline(transaction=True) as p:
        set_1, set_2, get_1, get_2 = await (p.set("blah", "boo").set("foo", "bar").get("blah").get("foo").execute())

    # response from redis.set is OK if successfully pushed
    assert set_1 is True
    assert set_2 is True
    assert get_1.decode() == "boo"
    assert get_2.decode() == "bar"


@pytest.mark.asyncio
@pytest.mark.snapshot(variants={"": aioredis_version >= (2, 0), "13": aioredis_version < (2, 0)})
async def test_two_traced_pipelines(redis_client):

    with tracer.trace("web-request", service="test"):
        if aioredis_version >= (2, 0):
            p1 = await redis_client.pipeline(transaction=False)
            p2 = await redis_client.pipeline(transaction=False)
            await p1.set("blah", "boo")
            await p2.set("foo", "bar")
            await p1.get("blah")
            await p2.get("foo")
        else:
            p1 = redis_client.pipeline()
            p2 = redis_client.pipeline()
            p1.set("blah", "boo")
            p2.set("foo", "bar")
            p1.get("blah")
            p2.get("foo")

        response_list1 = await p1.execute()
        response_list2 = await p2.execute()

    assert response_list1[0] is True  # response from redis.set is OK if successfully pushed
    assert response_list2[0] is True
    assert (
        response_list1[1].decode() == "boo"
    )  # response from hset is 'Integer reply: The number of fields that were added.'
    assert response_list2[1].decode() == "bar"


@pytest.mark.asyncio
@pytest.mark.snapshot(variants={"": aioredis_version >= (2, 0), "13": aioredis_version < (2, 0)})
async def test_parenting(redis_client):
    with tracer.trace("web-request", service="test"):
        await redis_client.set("blah", "boo")
        await redis_client.get("blah")



import collections
from copy import deepcopy
from typing import Any
from typing import Callable
from typing import DefaultDict
from typing import Optional
from typing import Set

import attr

from .internal.logger import get_logger


log = get_logger(__name__)


@attr.s(slots=True)
class Hooks(object):
    """
    Hooks configuration object is used for registering and calling hook functions

    Example::

        @config.falcon.hooks.on('request')
        def on_request(span, request, response):
            pass
    """

    _hooks = attr.ib(init=False, factory=lambda: collections.defaultdict(set), type=DefaultDict[str, Set])

    def __deepcopy__(self, memodict=None):
        hooks = Hooks()
        hooks._hooks = deepcopy(self._hooks, memodict)
        return hooks

    def register(
        self,
        hook,  # type: Any
        func=None,  # type: Optional[Callable]
    ):
        # type: (...) -> Optional[Callable[..., Any]]
        """
        Function used to register a hook for the provided name.

        Example::

            def on_request(span, request, response):
                pass

            config.falcon.hooks.register('request', on_request)


        If no function is provided then a decorator is returned::

            @config.falcon.hooks.register('request')
            def on_request(span, request, response):
                pass

        :param hook: The name of the hook to register the function for
        :type hook: object
        :param func: The function to register, or ``None`` if a decorator should be returned
        :type func: function, None
        :returns: Either a function decorator if ``func is None``, otherwise ``None``
        :rtype: function, None
        """
        # If they didn't provide a function, then return a decorator
        if not func:

            def wrapper(func):
                self.register(hook, func)
                return func

            return wrapper
        self._hooks[hook].add(func)
        return None

    # Provide shorthand `on` method for `register`
    # >>> @config.falcon.hooks.on('request')
    #     def on_request(span, request, response):
    #        pass
    on = register

    def deregister(
        self,
        hook,  # type: Any
        func,  # type: Callable
    ):
        # type: (...) -> None
        """
        Function to deregister a function from a hook it was registered under

        Example::

            @config.falcon.hooks.on('request')
            def on_request(span, request, response):
                pass

            config.falcon.hooks.deregister('request', on_request)

        :param hook: The name of the hook to register the function for
        :type hook: object
        :param func: Function hook to register
        :type func: function
        """
        if hook in self._hooks:
            try:
                self._hooks[hook].remove(func)
            except KeyError:
                pass

    def emit(
        self,
        hook,  # type: Any
        *args,  # type: Any
        **kwargs  # type: Any
    ):
        # type: (...) -> None
        """
        Function used to call registered hook functions.

        :param hook: The hook to call functions for
        :type hook: str
        :param args: Positional arguments to pass to the hook functions
        :type args: list
        :param kwargs: Keyword arguments to pass to the hook functions
        :type kwargs: dict
        """
        # Call registered hooks
        for func in self._hooks.get(hook, ()):
            try:
                func(*args, **kwargs)
            except Exception:
                log.error("Failed to run hook %s function %s", hook, func, exc_info=True)



import os

import psycopg2
from psycopg2.sql import Composable

from ddtrace import Pin
from ddtrace import config
from ddtrace.constants import SPAN_MEASURED_KEY
from ddtrace.contrib import dbapi
from ddtrace.contrib.trace_utils import ext_service
from ddtrace.ext import SpanTypes
from ddtrace.ext import db
from ddtrace.ext import net
from ddtrace.ext import sql
from ddtrace.vendor import wrapt

from ...internal.utils.formats import asbool
from ...internal.utils.version import parse_version


config._add(
    "psycopg",
    dict(
        _default_service="postgres",
        _dbapi_span_name_prefix="postgres",
        trace_fetch_methods=asbool(os.getenv("DD_PSYCOPG_TRACE_FETCH_METHODS", default=False)),
        trace_connect=asbool(os.getenv("DD_PSYCOPG_TRACE_CONNECT", default=False)),
    ),
)

# Original connect method
_connect = psycopg2.connect

PSYCOPG2_VERSION = parse_version(psycopg2.__version__)


def patch():
    """Patch monkey patches psycopg's connection function
    so that the connection's functions are traced.
    """
    if getattr(psycopg2, "_datadog_patch", False):
        return
    setattr(psycopg2, "_datadog_patch", True)

    Pin().onto(psycopg2)
    wrapt.wrap_function_wrapper(psycopg2, "connect", patched_connect)
    _patch_extensions(_psycopg2_extensions)  # do this early just in case


def unpatch():
    if getattr(psycopg2, "_datadog_patch", False):
        setattr(psycopg2, "_datadog_patch", False)
        psycopg2.connect = _connect
        _unpatch_extensions(_psycopg2_extensions)

        pin = Pin.get_from(psycopg2)
        if pin:
            pin.remove_from(psycopg2)


class Psycopg2TracedCursor(dbapi.TracedCursor):
    """TracedCursor for psycopg2"""

    def _trace_method(self, method, name, resource, extra_tags, *args, **kwargs):
        # treat psycopg2.sql.Composable resource objects as strings
        if isinstance(resource, Composable):
            resource = resource.as_string(self.__wrapped__)

        return super(Psycopg2TracedCursor, self)._trace_method(method, name, resource, extra_tags, *args, **kwargs)


class Psycopg2FetchTracedCursor(Psycopg2TracedCursor, dbapi.FetchTracedCursor):
    """FetchTracedCursor for psycopg2"""


class Psycopg2TracedConnection(dbapi.TracedConnection):
    """TracedConnection wraps a Connection with tracing code."""

    def __init__(self, conn, pin=None, cursor_cls=None):
        if not cursor_cls:
            # Do not trace `fetch*` methods by default
            cursor_cls = Psycopg2FetchTracedCursor if config.psycopg.trace_fetch_methods else Psycopg2TracedCursor

        super(Psycopg2TracedConnection, self).__init__(conn, pin, config.psycopg, cursor_cls=cursor_cls)


def patch_conn(conn, traced_conn_cls=Psycopg2TracedConnection):
    """Wrap will patch the instance so that its queries are traced."""
    # ensure we've patched extensions (this is idempotent) in
    # case we're only tracing some connections.
    _patch_extensions(_psycopg2_extensions)

    c = traced_conn_cls(conn)

    # fetch tags from the dsn
    dsn = sql.parse_pg_dsn(conn.dsn)
    tags = {
        net.TARGET_HOST: dsn.get("host"),
        net.TARGET_PORT: dsn.get("port"),
        db.NAME: dsn.get("dbname"),
        db.USER: dsn.get("user"),
        "db.application": dsn.get("application_name"),
    }

    Pin(tags=tags).onto(c)

    return c


def _patch_extensions(_extensions):
    # we must patch extensions all the time (it's pretty harmless) so split
    # from global patching of connections. must be idempotent.
    for _, module, func, wrapper in _extensions:
        if not hasattr(module, func) or isinstance(getattr(module, func), wrapt.ObjectProxy):
            continue
        wrapt.wrap_function_wrapper(module, func, wrapper)


def _unpatch_extensions(_extensions):
    # we must patch extensions all the time (it's pretty harmless) so split
    # from global patching of connections. must be idempotent.
    for original, module, func, _ in _extensions:
        setattr(module, func, original)


#
# monkeypatch targets
#


def patched_connect(connect_func, _, args, kwargs):
    pin = Pin.get_from(psycopg2)

    if not pin or not pin.enabled() or not config.psycopg.trace_connect:
        conn = connect_func(*args, **kwargs)
    else:
        with pin.tracer.trace(
            "psycopg2.connect", service=ext_service(pin, config.psycopg), span_type=SpanTypes.SQL
        ) as span:
            span.set_tag(SPAN_MEASURED_KEY)
            conn = connect_func(*args, **kwargs)
    return patch_conn(conn)


def _extensions_register_type(func, _, args, kwargs):
    def _unroll_args(obj, scope=None):
        return obj, scope

    obj, scope = _unroll_args(*args, **kwargs)

    # register_type performs a c-level check of the object
    # type so we must be sure to pass in the actual db connection
    if scope and isinstance(scope, wrapt.ObjectProxy):
        scope = scope.__wrapped__

    return func(obj, scope) if scope else func(obj)


def _extensions_quote_ident(func, _, args, kwargs):
    def _unroll_args(obj, scope=None):
        return obj, scope

    obj, scope = _unroll_args(*args, **kwargs)

    # register_type performs a c-level check of the object
    # type so we must be sure to pass in the actual db connection
    if scope and isinstance(scope, wrapt.ObjectProxy):
        scope = scope.__wrapped__

    return func(obj, scope) if scope else func(obj)


def _extensions_adapt(func, _, args, kwargs):
    adapt = func(*args, **kwargs)
    if hasattr(adapt, "prepare"):
        return AdapterWrapper(adapt)
    return adapt


class AdapterWrapper(wrapt.ObjectProxy):
    def prepare(self, *args, **kwargs):
        func = self.__wrapped__.prepare
        if not args:
            return func(*args, **kwargs)
        conn = args[0]

        # prepare performs a c-level check of the object type so
        # we must be sure to pass in the actual db connection
        if isinstance(conn, wrapt.ObjectProxy):
            conn = conn.__wrapped__

        return func(conn, *args[1:], **kwargs)


# extension hooks
_psycopg2_extensions = [
    (psycopg2.extensions.register_type, psycopg2.extensions, "register_type", _extensions_register_type),
    (psycopg2._psycopg.register_type, psycopg2._psycopg, "register_type", _extensions_register_type),
    (psycopg2.extensions.adapt, psycopg2.extensions, "adapt", _extensions_adapt),
]

# `_json` attribute is only available for psycopg >= 2.5
if getattr(psycopg2, "_json", None):
    _psycopg2_extensions += [
        (psycopg2._json.register_type, psycopg2._json, "register_type", _extensions_register_type),
    ]

# `quote_ident` attribute is only available for psycopg >= 2.7
if getattr(psycopg2, "extensions", None) and getattr(psycopg2.extensions, "quote_ident", None):
    _psycopg2_extensions += [
        (psycopg2.extensions.quote_ident, psycopg2.extensions, "quote_ident", _extensions_quote_ident),
    ]



import typing

import attr


if typing.TYPE_CHECKING:
    from .. import recorder


class ExportError(Exception):
    pass


@attr.s
class Exporter(object):
    """Exporter base class."""

    def export(
        self,
        events,  # type: recorder.EventsType
        start_time_ns,  # type: int
        end_time_ns,  # type: int
    ):
        # type: (...) -> typing.Any
        """Export events.

        :param events: List of events to export.
        :param start_time_ns: The start time of recording.
        :param end_time_ns: The end time of recording.
        """
        raise NotImplementedError


@attr.s
class NullExporter(Exporter):
    """Exporter that does nothing."""

    def export(
        self,
        events,  # type: recorder.EventsType
        start_time_ns,  # type: int
        end_time_ns,  # type: int
    ):
        # type: (...) -> None
        """Discard events."""
        pass



from cpython cimport *
from cpython.bytearray cimport PyByteArray_CheckExact
from libc cimport stdint
from libc.string cimport strlen

import threading

from ._utils cimport PyBytesLike_Check


# Do not use an absolute import here Cython<3.0.0 will
#   import `ddtrace.internal.constants` instead when this
#   package is installed in editable mode
# See the following for more details
#   https://github.com/DataDog/dd-trace-py/pull/4085
#   https://github.com/brettlangdon/shadow-import-issue
# DEV: This only occurs because there is a `constants.py` module
#   in both `ddtrace` and `ddtrace.internal`

from ..constants import ORIGIN_KEY


DEF MSGPACK_ARRAY_LENGTH_PREFIX_SIZE = 5
DEF MSGPACK_STRING_TABLE_LENGTH_PREFIX_SIZE = 6


cdef extern from "Python.h":
    const char* PyUnicode_AsUTF8(object o)

cdef extern from "pack.h":
    struct msgpack_packer:
        char* buf
        size_t length
        size_t buf_size

    int msgpack_pack_nil(msgpack_packer* pk)
    int msgpack_pack_long(msgpack_packer* pk, long d)
    int msgpack_pack_long_long(msgpack_packer* pk, long long d)
    int msgpack_pack_unsigned_long_long(msgpack_packer* pk, unsigned long long d)
    int msgpack_pack_double(msgpack_packer* pk, double d)
    int msgpack_pack_array(msgpack_packer* pk, size_t l)
    int msgpack_pack_map(msgpack_packer* pk, size_t l)
    int msgpack_pack_raw(msgpack_packer* pk, size_t l)
    int msgpack_pack_bin(msgpack_packer* pk, size_t l)
    int msgpack_pack_raw_body(msgpack_packer* pk, char* body, size_t l)
    int msgpack_pack_unicode(msgpack_packer* pk, object o, long long limit)
    int msgpack_pack_uint32(msgpack_packer* pk, stdint.uint32_t d)
    int msgpack_pack_uint64(msgpack_packer* pk, stdint.uint64_t d)
    int msgpack_pack_int32(msgpack_packer* pk, stdint.int32_t d)
    int msgpack_pack_int64(msgpack_packer* pk, stdint.int64_t d)
    int msgpack_pack_true(msgpack_packer* pk)
    int msgpack_pack_false(msgpack_packer* pk)


cdef long long ITEM_LIMIT = (2**32)-1


cdef inline int PyBytesLike_CheckExact(object o):
    return PyBytes_CheckExact(o) or PyByteArray_CheckExact(o)


class BufferFull(Exception):
    pass


class BufferItemTooLarge(Exception):
    pass


cdef inline const char * string_to_buff(str s):
    IF PY_MAJOR_VERSION >= 3:
        return PyUnicode_AsUTF8(s)
    ELSE:
        return <const char *> s


# This is a borrowed reference but should be fine as we don't expect ORIGIN_KEY
# to get GC'd.
cdef const char * _ORIGIN_KEY = string_to_buff(ORIGIN_KEY)
cdef size_t _ORIGIN_KEY_LEN = <size_t> len(ORIGIN_KEY)


cdef inline int array_prefix_size(stdint.uint32_t l):
    if l < 16:
        return 1
    elif l < (2<<16):
        return 3
    return MSGPACK_ARRAY_LENGTH_PREFIX_SIZE


cdef inline int pack_bytes(msgpack_packer *pk, char *bs, Py_ssize_t l):
    cdef int ret

    ret = msgpack_pack_raw(pk, l)
    if ret == 0:
        ret = msgpack_pack_raw_body(pk, bs, l)
    return ret


cdef inline int pack_number(msgpack_packer *pk, object n) except? -1:
    if n is None:
        return msgpack_pack_nil(pk)

    if PyLong_Check(n):
        # PyInt_Check(long) is True for Python 3.
        # So we should test long before int.
        try:
            if n > 0:
                return msgpack_pack_unsigned_long_long(pk, <unsigned long long> n)
            return msgpack_pack_long_long(pk, <long long> n)
        except OverflowError as oe:
            raise OverflowError("Integer value out of range")

    if PyInt_Check(n):
        return msgpack_pack_long(pk, <long> n)

    if PyFloat_Check(n):
        return msgpack_pack_double(pk, <double> n)

    raise TypeError("Unhandled numeric type: %r" % type(n))


cdef inline int pack_text(msgpack_packer *pk, object text) except? -1:
    cdef Py_ssize_t L
    cdef int ret

    if text is None:
        return msgpack_pack_nil(pk)

    if PyBytesLike_Check(text):
        L = len(text)
        if L > ITEM_LIMIT:
            PyErr_Format(ValueError, b"%.200s object is too large", Py_TYPE(text).tp_name)
        ret = msgpack_pack_raw(pk, L)
        if ret == 0:
            ret = msgpack_pack_raw_body(pk, <char *> text, L)
        return ret

    if PyUnicode_Check(text):
        IF PY_MAJOR_VERSION >= 3:
            ret = msgpack_pack_unicode(pk, text, ITEM_LIMIT)
            if ret == -2:
                raise ValueError("unicode string is too large")
        ELSE:
            text = PyUnicode_AsEncodedString(text, "utf-8", NULL)
            L = len(text)
            if L > ITEM_LIMIT:
                raise ValueError("unicode string is too large")
            ret = msgpack_pack_raw(pk, L)
            if ret == 0:
                ret = msgpack_pack_raw_body(pk, <char *> text, L)
        return ret

    raise TypeError("Unhandled text type: %r" % type(text))


cdef class StringTable(object):
    cdef dict _table
    cdef stdint.uint32_t _next_id

    def __init__(self):
        self._table = {"": 0}
        self.insert("")
        self._next_id = 1

    cdef insert(self, object string):
        pass

    cdef stdint.uint32_t _index(self, object string):
        cdef stdint.uint32_t _id

        if string is None:
            return 0

        if PyDict_Contains(self._table, string):
            return PyLong_AsLong(<object>PyDict_GetItem(self._table, string))

        _id = self._next_id
        PyDict_SetItem(self._table, string, PyLong_FromLong(_id))
        self.insert(string)
        self._next_id += 1
        return _id

    cpdef index(self, object string):
        return self._index(string)

    cdef reset(self):
        PyDict_Clear(self._table)
        PyDict_SetItem(self._table, "", 0)
        self.insert("")
        self._next_id = 1

    def __len__(self):
        return PyLong_FromLong(self._next_id)

    def __contains__(self, object string):
        return PyBool_FromLong(PyDict_Contains(self._table, string))


cdef class ListStringTable(StringTable):
    cdef list _list

    def __init__(self):
        self._list = []
        super(ListStringTable, self).__init__()

    cdef insert(self, object string):
        PyList_Append(self._list, string)

    def __iter__(self):
        return iter(self._list)


cdef class MsgpackStringTable(StringTable):
    cdef msgpack_packer pk
    cdef int max_size
    cdef int _sp_len
    cdef stdint.uint32_t _sp_id
    cdef object _lock
    cdef size_t _reset_size

    def __init__(self, max_size):
        self.pk.buf_size = min(max_size, 1 << 20)
        self.pk.buf = <char*> PyMem_Malloc(self.pk.buf_size)
        if self.pk.buf == NULL:
            raise MemoryError("Unable to allocate internal buffer.")
        self.max_size = max_size
        self.pk.length = MSGPACK_STRING_TABLE_LENGTH_PREFIX_SIZE
        self._sp_len = 0
        self._lock = threading.RLock()
        super(MsgpackStringTable, self).__init__()

        assert self.index(ORIGIN_KEY) == 1
        self._reset_size = self.pk.length

    def __dealloc__(self):
        PyMem_Free(self.pk.buf)
        self.pk.buf = NULL

    cdef insert(self, object string):
        cdef int ret

        if self.pk.length + len(string) > self.max_size:
            raise ValueError(
                "Cannot insert '%s': string table is full (current size: %d, max size: %d)." % (
                    string, self.pk.length, self.max_size
                )
            )

        ret = pack_text(&self.pk, string)
        if ret != 0:
            raise RuntimeError("Failed to add string to msgpack string table")

    cdef savepoint(self):
        self._sp_len = self.pk.length
        self._sp_id = self._next_id

    cdef rollback(self):
        if self._sp_len > 0:
            self.pk.length = self._sp_len
            self._next_id = self._sp_id

    cdef get_bytes(self):
        cdef int ret;
        cdef stdint.uint32_t l = self._next_id
        cdef int offset = MSGPACK_STRING_TABLE_LENGTH_PREFIX_SIZE - array_prefix_size(l)
        cdef int old_pos = self.pk.length
        
        with self._lock:
            # Update table size prefix
            self.pk.length = offset
            ret = msgpack_pack_array(&self.pk, l)
            if ret:
                return None
            # Add root array size prefix
            self.pk.length = offset = offset - 1
            ret = msgpack_pack_array(&self.pk, 2)
            if ret:
                return None
            self.pk.length = old_pos
        
        return PyBytes_FromStringAndSize(self.pk.buf + offset, self.pk.length - offset)

    @property
    def size(self):
        with self._lock:
            return self.pk.length - MSGPACK_ARRAY_LENGTH_PREFIX_SIZE + array_prefix_size(self._next_id)

    cdef append_raw(self, long src, Py_ssize_t size):
        cdef int res
        with self._lock:
            assert self.size + size <= self.max_size
            res = msgpack_pack_raw_body(&self.pk, <char *>PyLong_AsLong(src), size)
            if res != 0:
                raise RuntimeError("Failed to append raw bytes to msgpack string table")

    cdef reset(self):
        StringTable.reset(self)
        assert self._next_id == 1

        PyDict_SetItem(self._table, ORIGIN_KEY, 1)
        self._next_id = 2
        self.pk.length = self._reset_size
        self._sp_len = 0

    cpdef flush(self):
        with self._lock:
            try:
                return self.get_bytes()
            finally:
                self.reset()


cdef class BufferedEncoder(object):
    content_type: str = None

    cdef public size_t max_size
    cdef public size_t max_item_size
    cdef object _lock

    def __cinit__(self, size_t max_size, size_t max_item_size):
        self.max_size = max_size
        self.max_item_size = max_item_size
        self._lock = threading.Lock()

    # ---- Abstract methods ----

    def put(self, item):
        raise NotImplementedError()

    def encode(self):
        raise NotImplementedError()


cdef class ListBufferedEncoder(BufferedEncoder):
    cdef list _buffer
    cdef Py_ssize_t _size

    def __cinit__(self, size_t max_size, size_t max_item_size):
        self._buffer = []
        self._size = 0

    def __len__(self):
        return len(self._buffer)

    @property
    def size(self):
        with self._lock:
            return self._size

    cpdef put(self, item):
        """Put an item to be serialized in the buffer."""
        cdef int item_len

        encoded_item = self.encode_item(item)
        item_len = len(encoded_item)

        if item_len > self.max_item_size or item_len > self.max_size:
            raise BufferItemTooLarge(item_len)

        with self._lock:
            if self._size + item_len <= self.max_size:
                self._buffer.append(encoded_item)
                self._size += item_len
            else:
                raise BufferFull(item_len)

    cpdef get(self):
        """Get a copy of the buffer and clear it."""
        with self._lock:
            try:
                return list(self._buffer)
            finally:
                self._buffer[:] = []
                self._size = 0

    def encode_item(self, item):
        raise NotImplementedError()


cdef class MsgpackEncoderBase(BufferedEncoder):
    content_type = "application/msgpack"

    cdef msgpack_packer pk
    cdef stdint.uint32_t _count

    def __cinit__(self, size_t max_size, size_t max_item_size):
        cdef int buf_size = 1024*1024
        self.pk.buf = <char*> PyMem_Malloc(buf_size)
        if self.pk.buf == NULL:
            raise MemoryError("Unable to allocate internal buffer.")

        self.max_size = max_size
        self.pk.buf_size = buf_size
        self.max_item_size = max_item_size if max_item_size < max_size else max_size
        self._lock = threading.RLock()
        self._reset_buffer()

    def __dealloc__(self):
        PyMem_Free(self.pk.buf)
        self.pk.buf = NULL

    def __len__(self):  # TODO: Use a better name?
        return self._count

    cpdef _decode(self, data):
        import msgpack
        if msgpack.version[:2] < (0, 6):
            return msgpack.unpackb(data)
        return msgpack.unpackb(data, raw=True)

    cdef _reset_buffer(self):
        self._count = 0
        self.pk.length = MSGPACK_ARRAY_LENGTH_PREFIX_SIZE  # Leave room for array length prefix

    cpdef encode(self):
        with self._lock:
            if not self._count:
                return None

            return self.flush()

    cdef inline int _update_array_len(self):
        """Update traces array size prefix"""
        cdef int offset = MSGPACK_ARRAY_LENGTH_PREFIX_SIZE - array_prefix_size(self._count)
        cdef int old_pos = self.pk.length

        with self._lock:
            self.pk.length = offset
            msgpack_pack_array(&self.pk, self._count)
            self.pk.length = old_pos
            return offset

    cdef get_bytes(self):
        """Return internal buffer contents as bytes object"""
        cdef int offset = self._update_array_len()
        with self._lock:
            return PyBytes_FromStringAndSize(self.pk.buf + offset, self.pk.length - offset)

    cdef char * get_buffer(self):
        """Return internal buffer."""
        return self.pk.buf + self._update_array_len()

    cdef void * get_dd_origin_ref(self, str dd_origin):
        raise NotImplementedError()

    cdef inline int _pack_trace(self, list trace) except? -1:
        cdef int ret
        cdef Py_ssize_t L
        cdef void * dd_origin = NULL

        L = len(trace)
        if L > ITEM_LIMIT:
            raise ValueError("list is too large")

        ret = msgpack_pack_array(&self.pk, L)
        if ret != 0: raise RuntimeError("Couldn't pack trace")

        if L > 0 and trace[0].context is not None and trace[0].context.dd_origin is not None:
            dd_origin = self.get_dd_origin_ref(trace[0].context.dd_origin)

        for span in trace:
            ret = self.pack_span(span, dd_origin)
            if ret != 0: raise RuntimeError("Couldn't pack span")

        return ret

    cpdef put(self, list trace):
        """Put a trace (i.e. a list of spans) in the buffer."""
        cdef int ret

        with self._lock:
            len_before = self.pk.length
            size_before = self.size
            try:
                ret = self._pack_trace(trace)
                if ret:  # should not happen.
                    raise RuntimeError("internal error")

                # DEV: msgpack avoids buffer overflows by calling PyMem_Realloc so
                # we must check sizes manually.
                # TODO: We should probably ensure that the buffer size doesn't
                # grow arbitrarily because of the PyMem_Realloc and if it does then
                # free and reallocate with the appropriate size.
                if self.size - size_before > self.max_item_size:
                    raise BufferItemTooLarge(self.size - size_before)

                if self.size > self.max_size:
                    raise BufferFull(self.size - size_before)

                self._count += 1
            except:
                # rollback
                self.pk.length = len_before
                raise

    @property
    def size(self):
        """Return the size in bytes of the encoder buffer."""
        with self._lock:
            return self.pk.length + array_prefix_size(self._count) - MSGPACK_ARRAY_LENGTH_PREFIX_SIZE

    # ---- Abstract methods ----

    cpdef flush(self):
        raise NotImplementedError()

    cdef int pack_span(self, object span, void *dd_origin) except? -1:
        raise NotImplementedError()


cdef class MsgpackEncoderV03(MsgpackEncoderBase):
    cpdef flush(self):
        with self._lock:
            try:
                return self.get_bytes()
            finally:
                self._reset_buffer()

    cdef void * get_dd_origin_ref(self, str dd_origin):
        return string_to_buff(dd_origin)

    cdef inline int _pack_meta(self, object meta, char *dd_origin) except? -1:
        cdef Py_ssize_t L
        cdef int ret
        cdef dict d

        if PyDict_CheckExact(meta):
            d = <dict> meta
            L = len(d)
            if dd_origin is not NULL:
                L += 1
            if L > ITEM_LIMIT:
                raise ValueError("dict is too large")

            ret = msgpack_pack_map(&self.pk, L)
            if ret == 0:
                for k, v in d.items():
                    ret = pack_text(&self.pk, k)
                    if ret != 0: break
                    ret = pack_text(&self.pk, v)
                    if ret != 0: break
                if dd_origin is not NULL:
                    ret = pack_bytes(&self.pk, _ORIGIN_KEY, _ORIGIN_KEY_LEN)
                    if ret == 0:
                        ret = pack_bytes(&self.pk, dd_origin, strlen(dd_origin))
            return ret

        raise TypeError("Unhandled meta type: %r" % type(meta))

    cdef inline int _pack_metrics(self, object metrics) except? -1:
        cdef Py_ssize_t L
        cdef int ret
        cdef dict d

        if PyDict_CheckExact(metrics):
            d = <dict> metrics
            L = len(d)
            if L > ITEM_LIMIT:
                raise ValueError("dict is too large")

            ret = msgpack_pack_map(&self.pk, L)
            if ret == 0:
                for k, v in d.items():
                    ret = pack_text(&self.pk, k)
                    if ret != 0: break
                    ret = pack_number(&self.pk, v)
                    if ret != 0: break
            return ret

        raise TypeError("Unhandled metrics type: %r" % type(metrics))

    cdef int pack_span(self, object span, void *dd_origin) except? -1:
        cdef int ret
        cdef Py_ssize_t L
        cdef int has_span_type
        cdef int has_meta
        cdef int has_metrics

        has_error = <bint> (span.error != 0)
        has_span_type = <bint> (span.span_type is not None)
        has_meta = <bint> (len(span._meta) > 0 or dd_origin is not NULL)
        has_metrics = <bint> (len(span._metrics) > 0)
        has_parent_id = <bint> (span.parent_id is not None)

        L = 7 + has_span_type + has_meta + has_metrics + has_error + has_parent_id

        ret = msgpack_pack_map(&self.pk, L)

        if ret == 0:
            ret = pack_bytes(&self.pk, <char *> b"trace_id", 8)
            if ret != 0: return ret
            ret = pack_number(&self.pk, span.trace_id)
            if ret != 0: return ret

            if has_parent_id:
                ret = pack_bytes(&self.pk, <char *> b"parent_id", 9)
                if ret != 0: return ret
                ret = pack_number(&self.pk, span.parent_id)
                if ret != 0: return ret

            ret = pack_bytes(&self.pk, <char *> b"span_id", 7)
            if ret != 0: return ret
            ret = pack_number(&self.pk, span.span_id)
            if ret != 0: return ret

            ret = pack_bytes(&self.pk, <char *> b"service", 7)
            if ret != 0: return ret
            ret = pack_text(&self.pk, span.service)
            if ret != 0: return ret

            ret = pack_bytes(&self.pk, <char *> b"resource", 8)
            if ret != 0: return ret
            ret = pack_text(&self.pk, span.resource)
            if ret != 0: return ret

            ret = pack_bytes(&self.pk, <char *> b"name", 4)
            if ret != 0: return ret
            ret = pack_text(&self.pk, span.name)
            if ret != 0: return ret

            ret = pack_bytes(&self.pk, <char *> b"start", 5)
            if ret != 0: return ret
            ret = pack_number(&self.pk, span.start_ns)
            if ret != 0: return ret

            ret = pack_bytes(&self.pk, <char *> b"duration", 8)
            if ret != 0: return ret
            ret = pack_number(&self.pk, span.duration_ns)
            if ret != 0: return ret

            if has_error:
                ret = pack_bytes(&self.pk, <char *> b"error", 5)
                if ret != 0: return ret
                ret = msgpack_pack_long(&self.pk, <long> 1)
                if ret != 0: return ret

            if has_span_type:
                ret = pack_bytes(&self.pk, <char *> b"type", 4)
                if ret != 0: return ret
                ret = pack_text(&self.pk, span.span_type)
                if ret != 0: return ret

            if has_meta:
                ret = pack_bytes(&self.pk, <char *> b"meta", 4)
                if ret != 0: return ret
                ret = self._pack_meta(span._meta, <char *> dd_origin)
                if ret != 0: return ret

            if has_metrics:
                ret = pack_bytes(&self.pk, <char *> b"metrics", 7)
                if ret != 0: return ret
                ret = self._pack_metrics(span._metrics)
                if ret != 0: return ret

        return ret


cdef class MsgpackEncoderV05(MsgpackEncoderBase):
    cdef MsgpackStringTable _st

    def __cinit__(self, size_t max_size, size_t max_item_size):
        self._st = MsgpackStringTable(max_size)

    cpdef flush(self):
        with self._lock:
            try:
                self._st.append_raw(PyLong_FromLong(<long> self.get_buffer()), <Py_ssize_t> super(MsgpackEncoderV05, self).size)
                return self._st.flush()
            finally:
                self._reset_buffer()

    @property
    def size(self):
        """Return the size in bytes of the encoder buffer."""
        with self._lock:
            return self._st.size + super(MsgpackEncoderV05, self).size

    cpdef put(self, list trace):
        with self._lock:
            try:
                self._st.savepoint()
                super(MsgpackEncoderV05, self).put(trace)
            except Exception:
                self._st.rollback()
                raise

    cdef inline int _pack_string(self, object string):
        return msgpack_pack_uint32(&self.pk, self._st._index(string))

    cdef void * get_dd_origin_ref(self, str dd_origin):
        return <void *> PyLong_AsLong(self._st._index(dd_origin))

    cdef int pack_span(self, object span, void *dd_origin) except? -1:
        cdef int ret

        ret = msgpack_pack_array(&self.pk, 12)
        if ret != 0: return ret

        ret = self._pack_string(span.service)
        if ret != 0: return ret
        ret = self._pack_string(span.name)
        if ret != 0: return ret
        ret = self._pack_string(span.resource)
        if ret != 0: return ret

        _ = span.trace_id
        ret = msgpack_pack_uint64(&self.pk, _ if _ is not None else 0)
        if ret != 0: return ret
        
        _ = span.span_id
        ret = msgpack_pack_uint64(&self.pk, _ if _ is not None else 0)
        if ret != 0: return ret
        
        _ = span.parent_id
        ret = msgpack_pack_uint64(&self.pk, _ if _ is not None else 0)
        if ret != 0: return ret
        
        _ = span.start_ns
        ret = msgpack_pack_int64(&self.pk, _ if _ is not None else 0)
        if ret != 0: return ret
        
        _ = span.duration_ns
        ret = msgpack_pack_int64(&self.pk, _ if _ is not None else 0)
        if ret != 0: return ret
        
        _ = span.error
        ret = msgpack_pack_int32(&self.pk, _ if _ is not None else 0)
        if ret != 0: return ret

        ret = msgpack_pack_map(&self.pk, len(span._meta) + (dd_origin is not NULL))
        if ret != 0: return ret
        if span._meta:
            for k, v in span._meta.items():
                ret = self._pack_string(k)
                if ret != 0: return ret
                ret = self._pack_string(v)
                if ret != 0: return ret
        if dd_origin is not NULL:
            ret = msgpack_pack_uint32(&self.pk, <stdint.uint32_t> 1)
            if ret != 0: return ret
            ret = msgpack_pack_uint32(&self.pk, <stdint.uint32_t> dd_origin)
            if ret != 0: return ret
        
        ret = msgpack_pack_map(&self.pk, len(span._metrics))
        if ret != 0: return ret
        if span._metrics:
            for k, v in span._metrics.items():
                ret = self._pack_string(k)
                if ret != 0: return ret
                ret = pack_number(&self.pk, v)
                if ret != 0: return ret

        ret = self._pack_string(span.span_type)
        if ret != 0: return ret

        return 0


cdef class Packer(object):
    """Slightly modified version of the v0.6.2 msgpack Packer
    which only supports basic Python types (int, bool, float, dict, list).

    Note that _only_ the basic types can be encoded. Subtypes of these types
    are not supported.

    - strict_type argument is removed and assumed to be True
    - use_bin_type argument is removed and assumed to be True (use the msgpack 2.0 bin type fields when possible)
    - use_single_float is removed and assumed to be False
    - autoreset is removed and assumed to be True (bytes are always returned from pack and the buffer reset)

    https://github.com/msgpack/msgpack-python/tree/v0.6.2
    """
    cdef msgpack_packer pk
    cdef object _default
    cdef object _berrors
    cdef const char *encoding
    cdef const char *unicode_errors

    def __cinit__(self):
        cdef int buf_size = 1024*1024
        self.pk.buf = <char*> PyMem_Malloc(buf_size)
        if self.pk.buf == NULL:
            raise MemoryError("Unable to allocate internal buffer.")
        self.pk.buf_size = buf_size
        self.pk.length = 0

    def __init__(self, default=None):
        if default is not None:
            if not PyCallable_Check(default):
                raise TypeError("default must be a callable.")
        self._default = default

        if PY_MAJOR_VERSION < 3:
            self.encoding = "utf-8"
        else:
            self.encoding = NULL

    def __dealloc__(self):
        PyMem_Free(self.pk.buf)
        self.pk.buf = NULL

    cdef int _pack(self, object o) except -1:
        cdef long long llval
        cdef unsigned long long ullval
        cdef long longval
        cdef float fval
        cdef double dval
        cdef char* rawval
        cdef int ret
        cdef dict d
        cdef Py_ssize_t L
        cdef int default_used = 0
        cdef Py_buffer view
        cdef long i

        while True:
            if o is None:
                ret = msgpack_pack_nil(&self.pk)
            elif PyLong_CheckExact(o):
                # PyInt_Check(long) is True for Python 3.
                # So we should test long before int.
                try:
                    if o > 0:
                        ullval = o
                        ret = msgpack_pack_unsigned_long_long(&self.pk, ullval)
                    else:
                        llval = o
                        ret = msgpack_pack_long_long(&self.pk, llval)
                except OverflowError as oe:
                    if not default_used and self._default is not None:
                        o = self._default(o)
                        default_used = True
                        continue
                    else:
                        raise OverflowError("Integer value out of range")
            elif PyInt_CheckExact(o):
                longval = o
                ret = msgpack_pack_long(&self.pk, longval)
            elif PyFloat_CheckExact(o):
                dval = o
                ret = msgpack_pack_double(&self.pk, dval)
            elif PyBytesLike_CheckExact(o):
                L = len(o)
                if L > ITEM_LIMIT:
                    PyErr_Format(ValueError, b"%.200s object is too large", Py_TYPE(o).tp_name)
                rawval = o
                ret = msgpack_pack_bin(&self.pk, L)
                if ret == 0:
                    ret = msgpack_pack_raw_body(&self.pk, rawval, L)
            elif PyUnicode_CheckExact(o):
                if self.encoding == NULL:
                    ret = msgpack_pack_unicode(&self.pk, o, ITEM_LIMIT)
                    if ret == -2:
                        raise ValueError("unicode string is too large")
                else:
                    o = PyUnicode_AsEncodedString(o, self.encoding, self.unicode_errors)
                    L = len(o)
                    if L > ITEM_LIMIT:
                        raise ValueError("unicode string is too large")
                    ret = msgpack_pack_raw(&self.pk, L)
                    if ret == 0:
                        rawval = o
                        ret = msgpack_pack_raw_body(&self.pk, rawval, L)
            elif PyDict_CheckExact(o):
                d = <dict>o
                L = len(d)
                if L > ITEM_LIMIT:
                    raise ValueError("dict is too large")
                ret = msgpack_pack_map(&self.pk, L)
                if ret == 0:
                    for k, v in d.items():
                       ret = self._pack(k)
                       if ret != 0: break
                       ret = self._pack(v)
                       if ret != 0: break
            elif PyList_CheckExact(o):
                L = Py_SIZE(o)
                if L > ITEM_LIMIT:
                    raise ValueError("list is too large")
                ret = msgpack_pack_array(&self.pk, L)
                if ret == 0:
                    for v in o:
                        ret = self._pack(v)
                        if ret != 0: break
            elif PyBool_Check(o):
                if o:
                    ret = msgpack_pack_true(&self.pk)
                else:
                    ret = msgpack_pack_false(&self.pk)
            else:
                PyErr_Format(TypeError, b"can not serialize '%.200s' object", Py_TYPE(o).tp_name)
            return ret

    cpdef pack(self, object obj):
        cdef int ret
        try:
            ret = self._pack(obj)
        except:
            self.pk.length = 0

            raise
        if ret:  # should not happen.
            raise RuntimeError("internal error")

        # Reset the buffer.
        buf = PyBytes_FromStringAndSize(self.pk.buf, self.pk.length)
        self.pk.length = 0
        return buf

    def bytes(self):
        """Return internal buffer contents as bytes object"""
        return PyBytes_FromStringAndSize(self.pk.buf, self.pk.length)


def packb(o, **kwargs):
    """
    Pack object `o` and return packed bytes
    See :class:`Packer` for options.
    """
    return Packer(**kwargs).pack(o)



from ddtrace.internal.wrapping import unwrap
from ddtrace.internal.wrapping import wrap


def test_wrap_unwrap():
    channel = []

    def wrapper(f, args, kwargs):
        channel[:] = []
        channel.append((args, kwargs))
        retval = f(*args, **kwargs)
        channel.append(retval)
        return channel

    def f(a, b, c=None):
        return (a, b, c)

    wrap(f, wrapper)

    assert f(1, 2, 3) == [((1, 2, 3), {}), (1, 2, 3)]
    assert f(1, b=2, c=3) == [((1, 2, 3), {}), (1, 2, 3)]

    channel[:] = []
    unwrap(f, wrapper)
    assert f(1, 2, 3) == (1, 2, 3)
    assert not channel


def test_mutiple_wrap():
    channel1, channel2 = [], []

    def wrapper_maker(channel):
        def wrapper(f, args, kwargs):
            channel[:] = []
            channel.append((args, kwargs))
            retval = f(*args, **kwargs)
            channel.append(retval)
            return retval

        return wrapper

    wrapper1, wrapper2 = (wrapper_maker(_) for _ in (channel1, channel2))

    def f(a, b, c=None):
        return (a, b, c)

    wrap(f, wrapper1)
    wrap(f, wrapper2)

    f(1, 2, 3)

    assert channel1 == channel2 == [((1, 2, 3), {}), (1, 2, 3)]

    channel1[:] = []
    channel2[:] = []

    unwrap(f, wrapper1)
    f(1, 2, 3)

    assert not channel1 and channel2 == [((1, 2, 3), {}), (1, 2, 3)]

    channel2[:] = []

    unwrap(f, wrapper2)
    f(1, 2, 3)

    assert not channel1 and not channel2


def test_wrap_generator():
    channel = []

    def wrapper(f, args, kwargs):
        channel[:] = []
        for _ in f(*args, **kwargs):
            channel.append(_)
            yield _

    def g():
        for _ in range(10):
            yield _
        return

    wrap(g, wrapper)

    assert list(g()) == list(range(10)) == channel


def test_wrap_generator_send():
    def wrapper(f, args, kwargs):
        return f(*args, **kwargs)

    def g():
        yield 0
        for _ in range(1, 10):
            n = yield _
            assert _ == n
        return

    wrap(g, wrapper)

    gen = g()
    n = next(gen)
    channel = [n]
    try:
        while True:
            n = gen.send(n)
            channel.append(n)
    except StopIteration:
        pass

    assert list(range(10)) == channel



from collections import defaultdict
import copy

import bm as bm
import bm.utils as utils

from ddtrace import config as ddconfig
from ddtrace.contrib.trace_utils import set_http_meta


class Config(defaultdict):
    __header_tag_name = {
        "User-Agent": "http.user_agent",
        "REFERER": "http.referer",
        "Content-Type": "http.content_type",
        "Etag": "http.etag",
    }

    def _header_tag_name(self, header_name):
        return self.__header_tag_name.get(header_name)

    def __getattr__(self, item):
        return self[item]


COOKIES = {"csrftoken": "cR8TVoVebF2afssCR16pQeqHcxAlA3867P6zkkUBYDL5Q92kjSGtqptAry1htdlL"}

DATA_GET = dict(
    method="GET",
    status_code=200,
    status_msg="OK",
    parsed_query={
        "key1": "value1",
        "key2": "value2",
        "token": "cR8TVoVebF2afssCR16pQeqHcxAlA3867P6zkkUBYDL5Q92kjSGtqptAry1htdlL",
    },
    request_headers=utils.COMMON_DJANGO_META,
    response_headers=utils.COMMON_DJANGO_META,
    retries_remain=0,
    raw_uri="http://localhost:8888{}?key1=value1&key2=value2&token="
    "cR8TVoVebF2afssCR16pQeqHcxAlA3867P6zkkUBYDL5Q92kjSGtqptAry1htdlL".format(utils.PATH),
    request_cookies=COOKIES,
    request_path_params={"id": 1},
)


class SetHttpMeta(bm.Scenario):
    allenabled = bm.var_bool()
    useragentvariant = bm.var(type=str)
    obfuscation_disabled = bm.var_bool()
    send_querystring_enabled = bm.var_bool()
    url = bm.var(type=str)
    querystring = bm.var(type=str)
    ip_header = bm.var(type=str)
    ip_disabled = bm.var_bool()

    def run(self):
        # run scenario to also set tags on spans
        if self.allenabled:
            config = Config(lambda: True)
        else:
            config = Config(lambda: False)

        # querystring obfuscation config
        config["trace_query_string"] = self.send_querystring_enabled
        if self.obfuscation_disabled:
            ddconfig._obfuscation_query_string_pattern = None

        data = copy.deepcopy(DATA_GET)
        data["url"] = self.url
        data["query"] = self.querystring

        if self.useragentvariant:
            data["request_headers"][self.useragentvariant] = (
                "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 "
                "(KHTML, like Gecko) Chrome/92.0.4515.159 Safari/537.36"
            )

        if self.ip_header:
            data["request_headers"][self.ip_header] = "8.8.8.8"

        span = utils.gen_span(str("test"))
        span._local_root = utils.gen_span(str("root"))

        def bm(loops):
            with utils.override_env(
                dict(
                    DD_TRACE_CLIENT_IP_HEADER_DISABLED=str(self.ip_disabled),
                    DD_TRACE_CLIENT_IP_HEADER=self.ip_header,
                )
            ):
                for _ in range(loops):
                    set_http_meta(span, config, **data)

        yield bm



from .utils.cache import cachedmethod


class GlobMatcher(object):
    """This is a backtracking implementation of the glob matching algorithm.
    The glob pattern language supports `*` as a multiple character wildcard which includes matches on `""`
    and `?` as a single character wildcard, but no escape sequences.
    The match method will be cached for quicker matching and is in a class to keep it from being global.
    """

    def __init__(self, pattern):
        # type: (str) -> None
        self.pattern = pattern

    @cachedmethod()
    def match(self, subject):
        # type: (str) -> bool
        pattern = self.pattern
        px = 0  # [p]attern inde[x]
        sx = 0  # [s]ubject inde[x]
        nextPx = 0
        nextSx = 0

        while px < len(pattern) or sx < len(subject):
            if px < len(pattern):
                char = pattern[px]

                if char == "?":  # single character wildcard
                    if sx < len(subject):
                        px += 1
                        sx += 1
                        continue

                elif char == "*":  # zero-or-more-character wildcard
                    nextPx = px
                    nextSx = sx + 1
                    px += 1
                    continue

                elif sx < len(subject) and subject[sx] == char:  # default normal character match
                    px += 1
                    sx += 1
                    continue

            if 0 < nextSx <= len(subject):
                px = nextPx
                sx = nextSx
                continue

            return False
        return True



from unittest import TestCase

import pytest

from ddtrace import Pin


class PinTestCase(TestCase):
    """TestCase for the `Pin` object that is used when an object is wrapped
    with our tracing functionalities.
    """

    def setUp(self):
        # define a simple class object
        class Obj(object):
            pass

        self.Obj = Obj

    def test_pin(self):
        # ensure a Pin can be attached to an instance
        obj = self.Obj()
        pin = Pin(service="metrics")
        pin.onto(obj)

        got = Pin.get_from(obj)
        assert got.service == pin.service
        assert got is pin

    def test_pin_find(self):
        # ensure Pin will find the first available pin

        # Override service
        obj_a = self.Obj()
        pin = Pin(service="service-a")
        pin.onto(obj_a)

        # Override service
        obj_b = self.Obj()
        pin = Pin(service="service-b")
        pin.onto(obj_b)

        # No Pin set
        obj_c = self.Obj()

        # We find the first pin (obj_b)
        pin = Pin._find(obj_c, obj_b, obj_a)
        assert pin is not None
        assert pin.service == "service-b"

        # We find the first pin (obj_a)
        pin = Pin._find(obj_a, obj_b, obj_c)
        assert pin is not None
        assert pin.service == "service-a"

        # We don't find a pin if none is there
        pin = Pin._find(obj_c, obj_c, obj_c)
        assert pin is None

    def test_cant_pin_with_slots(self):
        # ensure a Pin can't be attached if the __slots__ is defined
        class Obj(object):
            __slots__ = ["value"]

        obj = Obj()
        obj.value = 1

        Pin(service="metrics").onto(obj)
        got = Pin.get_from(obj)
        assert got is None

    def test_cant_modify(self):
        # ensure a Pin is immutable once initialized
        pin = Pin(service="metrics")
        with pytest.raises(AttributeError):
            pin.service = "intake"

    def test_copy(self):
        # ensure a Pin is copied when using the clone methods
        p1 = Pin(service="metrics", tags={"key": "value"})
        p2 = p1.clone(service="intake")
        # values are the same
        assert p1.service == "metrics"
        assert p2.service == "intake"
        # but it's a copy
        assert p1.tags is not p2.tags
        assert p1._config is not p2._config
        # of almost everything
        assert p1.tracer is p2.tracer

    def test_none(self):
        # ensure get_from returns None if a Pin is not available
        assert Pin.get_from(None) is None

    def test_repr(self):
        # ensure the service name is in the string representation of the Pin
        pin = Pin(service="metrics")
        assert "metrics" in str(pin)

    def test_override(self):
        # ensure Override works for an instance object
        class A(object):
            pass

        Pin(service="metrics").onto(A)
        a = A()
        Pin.override(a)
        assert Pin.get_from(a).service == "metrics"

        b = A()
        assert Pin.get_from(b).service == "metrics"

    def test_override_missing(self):
        # ensure overriding an instance doesn't override the Class
        class A(object):
            pass

        a = A()
        assert Pin.get_from(a) is None
        Pin.override(a, service="metrics")
        assert Pin.get_from(a).service == "metrics"

        b = A()
        assert Pin.get_from(b) is None

    def test_pin_config(self):
        # ensure `Pin` has a configuration object that can be modified
        obj = self.Obj()
        Pin.override(obj, service="metrics")
        pin = Pin.get_from(obj)
        assert pin._config is not None
        pin._config["distributed_tracing"] = True
        assert pin._config["distributed_tracing"] is True

    def test_pin_config_is_a_copy(self):
        # ensure that when a `Pin` is cloned, the config is a copy
        obj = self.Obj()
        Pin.override(obj, service="metrics")
        p1 = Pin.get_from(obj)
        assert p1._config is not None
        p1._config["distributed_tracing"] = True

        Pin.override(obj, service="intake")
        p2 = Pin.get_from(obj)
        assert p2._config is not None
        p2._config["distributed_tracing"] = False

        assert p1._config["distributed_tracing"] is True
        assert p2._config["distributed_tracing"] is False

    def test_pin_does_not_override_global(self):
        # ensure that when a `Pin` is created from a class, the specific
        # instance doesn't override the global one
        class A(object):
            pass

        Pin.override(A, service="metrics")
        global_pin = Pin.get_from(A)
        global_pin._config["distributed_tracing"] = True

        a = A()
        pin = Pin.get_from(a)
        assert pin is not None
        assert pin._config["distributed_tracing"] is True
        pin._config["distributed_tracing"] = False

        assert global_pin._config["distributed_tracing"] is True
        assert pin._config["distributed_tracing"] is False

    def test_pin_does_not_override_global_with_new_instance(self):
        # ensure that when a `Pin` is created from a class, the specific
        # instance doesn't override the global one, even if only the
        # `onto()` API has been used
        class A(object):
            pass

        pin = Pin(service="metrics")
        pin.onto(A)
        global_pin = Pin.get_from(A)
        global_pin._config["distributed_tracing"] = True

        a = A()
        pin = Pin.get_from(a)
        assert pin is not None
        assert pin._config["distributed_tracing"] is True
        pin._config["distributed_tracing"] = False

        assert global_pin._config["distributed_tracing"] is True
        assert pin._config["distributed_tracing"] is False



# tags
CLUSTER = "cassandra.cluster"
KEYSPACE = "cassandra.keyspace"
CONSISTENCY_LEVEL = "cassandra.consistency_level"
PAGINATED = "cassandra.paginated"
ROW_COUNT = "cassandra.row_count"
PAGE_NUMBER = "cassandra.page_number"



import flask

from ddtrace import Pin
from ddtrace import config

from .. import trace_utils


def get_current_app():
    """Helper to get the flask.app.Flask from the current app context"""
    appctx = flask._app_ctx_stack.top
    if appctx:
        return appctx.app
    return None


def with_instance_pin(func):
    """Helper to wrap a function wrapper and ensure an enabled pin is available for the `instance`"""

    def wrapper(wrapped, instance, args, kwargs):
        pin = Pin._find(wrapped, instance, get_current_app())
        if not pin or not pin.enabled():
            return wrapped(*args, **kwargs)

        return func(pin, wrapped, instance, args, kwargs)

    return wrapper


def simple_tracer(name, span_type=None):
    """Generate a simple tracer that wraps the function call with `with tracer.trace()`"""

    @with_instance_pin
    def wrapper(pin, wrapped, instance, args, kwargs):
        with pin.tracer.trace(name, service=trace_utils.int_service(pin, config.flask, pin), span_type=span_type):
            return wrapped(*args, **kwargs)

    return wrapper



# Tests are copied from cpython/Lib/test/test_context.py
# License: PSFL
# Copyright: 2018 Python Software Foundation


import concurrent.futures
import functools
import random
import time
import unittest

import pytest

from ddtrace.vendor import contextvars


def isolated_context(func):
    """Needed to make reftracking test mode work."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        ctx = contextvars.Context()
        return ctx.run(func, *args, **kwargs)

    return wrapper


class ContextTest(unittest.TestCase):
    def test_context_var_new_1(self):
        with self.assertRaises(TypeError):
            contextvars.ContextVar()

        with pytest.raises(TypeError) as e:
            contextvars.ContextVar(1)
        assert "must be a str" in str(e.value)

        c = contextvars.ContextVar("a")
        self.assertNotEqual(hash(c), hash("a"))

    @isolated_context
    def test_context_var_repr_1(self):
        c = contextvars.ContextVar("a")
        self.assertIn("a", repr(c))

        c = contextvars.ContextVar("a", default=123)
        self.assertIn("123", repr(c))

        lst = []
        c = contextvars.ContextVar("a", default=lst)
        lst.append(c)
        self.assertIn("...", repr(c))
        self.assertIn("...", repr(lst))

        t = c.set(1)
        self.assertIn(repr(c), repr(t))
        self.assertNotIn(" used ", repr(t))
        c.reset(t)
        self.assertIn(" used ", repr(t))

    def test_context_new_1(self):
        with self.assertRaises(TypeError):
            contextvars.Context(1)
        with self.assertRaises(TypeError):
            contextvars.Context(1, a=1)
        with self.assertRaises(TypeError):
            contextvars.Context(a=1)
        contextvars.Context(**{})

    def test_context_typerrors_1(self):
        ctx = contextvars.Context()

        with pytest.raises(TypeError) as e:
            ctx[1]
        assert "ContextVar key was expected" in str(e.value)

        with pytest.raises(TypeError):
            1 in ctx
        assert "ContextVar key was expected" in str(e.value)

        with pytest.raises(TypeError) as e:
            ctx.get(1)
        assert "ContextVar key was expected" in str(e.value)

    def test_context_get_context_1(self):
        ctx = contextvars.copy_context()
        self.assertIsInstance(ctx, contextvars.Context)

    def test_context_run_1(self):
        ctx = contextvars.Context()

        with pytest.raises(TypeError):
            ctx.run()

    def test_context_run_2(self):
        ctx = contextvars.Context()

        def func(*args, **kwargs):
            kwargs["spam"] = "foo"
            args += ("bar",)
            return args, kwargs

        for f in (func, functools.partial(func)):
            # partial doesn't support FASTCALL

            self.assertEqual(ctx.run(f), (("bar",), {"spam": "foo"}))
            self.assertEqual(ctx.run(f, 1), ((1, "bar"), {"spam": "foo"}))

            self.assertEqual(ctx.run(f, a=2), (("bar",), {"a": 2, "spam": "foo"}))

            self.assertEqual(ctx.run(f, 11, a=2), ((11, "bar"), {"a": 2, "spam": "foo"}))

            a = {}
            self.assertEqual(ctx.run(f, 11, **a), ((11, "bar"), {"spam": "foo"}))
            self.assertEqual(a, {})

    def test_context_run_3(self):
        ctx = contextvars.Context()

        def func(*args, **kwargs):
            1 / 0

        with self.assertRaises(ZeroDivisionError):
            ctx.run(func)
        with self.assertRaises(ZeroDivisionError):
            ctx.run(func, 1, 2)
        with self.assertRaises(ZeroDivisionError):
            ctx.run(func, 1, 2, a=123)

    @isolated_context
    def test_context_run_4(self):
        ctx1 = contextvars.Context()
        ctx2 = contextvars.Context()
        var = contextvars.ContextVar("var")

        def func2():
            self.assertIsNone(var.get(None))

        def func1():
            self.assertIsNone(var.get(None))
            var.set("spam")
            ctx2.run(func2)
            self.assertEqual(var.get(None), "spam")

            cur = contextvars.copy_context()
            self.assertEqual(len(cur), 1)
            self.assertEqual(cur[var], "spam")
            return cur

        returned_ctx = ctx1.run(func1)
        self.assertEqual(ctx1, returned_ctx)
        self.assertEqual(returned_ctx[var], "spam")
        self.assertIn(var, returned_ctx)

    def test_context_run_5(self):
        ctx = contextvars.Context()
        var = contextvars.ContextVar("var")

        def func():
            self.assertIsNone(var.get(None))
            var.set("spam")
            1 / 0

        with self.assertRaises(ZeroDivisionError):
            ctx.run(func)

        self.assertIsNone(var.get(None))

    def test_context_run_6(self):
        ctx = contextvars.Context()
        c = contextvars.ContextVar("a", default=0)

        def fun():
            self.assertEqual(c.get(), 0)
            self.assertIsNone(ctx.get(c))

            c.set(42)
            self.assertEqual(c.get(), 42)
            self.assertEqual(ctx.get(c), 42)

        ctx.run(fun)

    def test_context_run_7(self):
        ctx = contextvars.Context()

        def fun():
            with pytest.raises(RuntimeError) as e:
                ctx.run(fun)
            assert "is already entered" in str(e.value)

        ctx.run(fun)

    @isolated_context
    def test_context_getset_1(self):
        c = contextvars.ContextVar("c")
        with self.assertRaises(LookupError):
            c.get()

        self.assertIsNone(c.get(None))

        t0 = c.set(42)
        self.assertEqual(c.get(), 42)
        self.assertEqual(c.get(None), 42)
        self.assertIs(t0.old_value, t0.MISSING)
        self.assertIs(t0.old_value, contextvars.Token.MISSING)
        self.assertIs(t0.var, c)

        t = c.set("spam")
        self.assertEqual(c.get(), "spam")
        self.assertEqual(c.get(None), "spam")
        self.assertEqual(t.old_value, 42)
        c.reset(t)

        self.assertEqual(c.get(), 42)
        self.assertEqual(c.get(None), 42)

        c.set("spam2")
        with pytest.raises(RuntimeError) as e:
            c.reset(t)
        assert "has already been used" in str(e.value)
        self.assertEqual(c.get(), "spam2")

        ctx1 = contextvars.copy_context()
        self.assertIn(c, ctx1)

        c.reset(t0)
        with pytest.raises(RuntimeError):
            c.reset(t0)
        assert "has already been used" in str(e.value)
        self.assertIsNone(c.get(None))

        self.assertIn(c, ctx1)
        self.assertEqual(ctx1[c], "spam2")
        self.assertEqual(ctx1.get(c, "aa"), "spam2")
        self.assertEqual(len(ctx1), 1)
        self.assertEqual(list(ctx1.items()), [(c, "spam2")])
        self.assertEqual(list(ctx1.values()), ["spam2"])
        self.assertEqual(list(ctx1.keys()), [c])
        self.assertEqual(list(ctx1), [c])

        ctx2 = contextvars.copy_context()
        self.assertNotIn(c, ctx2)
        with self.assertRaises(KeyError):
            ctx2[c]
        self.assertEqual(ctx2.get(c, "aa"), "aa")
        self.assertEqual(len(ctx2), 0)
        self.assertEqual(list(ctx2), [])

    @isolated_context
    def test_context_getset_2(self):
        v1 = contextvars.ContextVar("v1")
        v2 = contextvars.ContextVar("v2")

        t1 = v1.set(42)
        with pytest.raises(ValueError) as e:
            v2.reset(t1)
        assert "by a different" in str(e.value)

    @isolated_context
    def test_context_getset_3(self):
        c = contextvars.ContextVar("c", default=42)
        ctx = contextvars.Context()

        def fun():
            self.assertEqual(c.get(), 42)
            with self.assertRaises(KeyError):
                ctx[c]
            self.assertIsNone(ctx.get(c))
            self.assertEqual(ctx.get(c, "spam"), "spam")
            self.assertNotIn(c, ctx)
            self.assertEqual(list(ctx.keys()), [])

            t = c.set(1)
            self.assertEqual(list(ctx.keys()), [c])
            self.assertEqual(ctx[c], 1)

            c.reset(t)
            self.assertEqual(list(ctx.keys()), [])
            with self.assertRaises(KeyError):
                ctx[c]

        ctx.run(fun)

    @isolated_context
    def test_context_getset_4(self):
        c = contextvars.ContextVar("c", default=42)
        ctx = contextvars.Context()

        tok = ctx.run(c.set, 1)

        with pytest.raises(ValueError) as e:
            c.reset(tok)
        assert "different Context" in str(e.value)

    @isolated_context
    def test_context_getset_5(self):
        c = contextvars.ContextVar("c", default=42)
        c.set([])

        def fun():
            c.set([])
            c.get().append(42)
            self.assertEqual(c.get(), [42])

        contextvars.copy_context().run(fun)
        self.assertEqual(c.get(), [])

    def test_context_copy_1(self):
        ctx1 = contextvars.Context()
        c = contextvars.ContextVar("c", default=42)

        def ctx1_fun():
            c.set(10)

            ctx2 = ctx1.copy()
            self.assertEqual(ctx2[c], 10)

            c.set(20)
            self.assertEqual(ctx1[c], 20)
            self.assertEqual(ctx2[c], 10)

            ctx2.run(ctx2_fun)
            self.assertEqual(ctx1[c], 20)
            self.assertEqual(ctx2[c], 30)

        def ctx2_fun():
            self.assertEqual(c.get(), 10)
            c.set(30)
            self.assertEqual(c.get(), 30)

        ctx1.run(ctx1_fun)

    @isolated_context
    def test_context_threads_1(self):
        cvar = contextvars.ContextVar("cvar")

        def sub(num):
            for i in range(10):
                cvar.set(num + i)
                time.sleep(random.uniform(0.001, 0.05))
                self.assertEqual(cvar.get(), num + i)
            return num

        tp = concurrent.futures.ThreadPoolExecutor(max_workers=10)
        try:
            results = list(tp.map(sub, range(10)))
        finally:
            tp.shutdown()
        self.assertEqual(results, list(range(10)))



"""
ddtrace.vendor
==============
Install vendored dependencies under a different top level package to avoid importing `ddtrace/__init__.py`
whenever a dependency is imported. Doing this allows us to have a little more control over import order.


Dependencies
============

wrapt
-----

Website: https://wrapt.readthedocs.io/en/latest/
Source: https://github.com/GrahamDumpleton/wrapt/
Version: 1.14.1
License: BSD 2-Clause "Simplified" License

Notes:
  `setup.py` will attempt to build the `wrapt/_wrappers.c` C module


dogstatsd
---------

Website: https://datadogpy.readthedocs.io/en/latest/
Source: https://github.com/DataDog/datadogpy
Version: 8e11af2 (0.39.1)
License: Copyright (c) 2020, Datadog <info@datadoghq.com>

Notes:
  `dogstatsd/__init__.py` was updated to include a copy of the `datadogpy` license: https://github.com/DataDog/datadogpy/blob/master/LICENSE
  Only `datadog.dogstatsd` module was vendored to avoid unnecessary dependencies
  `datadog/util/compat.py` was copied to `dogstatsd/compat.py`
  `datadog/util/format.py` was copied to `dogstatsd/format.py`
  version fixed to 8e11af2
  removed type imports
  removed unnecessary compat utils


monotonic
---------

Website: https://pypi.org/project/monotonic/
Source: https://github.com/atdt/monotonic
Version: 1.5
License: Apache License 2.0

Notes:
  The source `monotonic.py` was added as `monotonic/__init__.py`

  No other changes were made

debtcollector
-------------

Website: https://docs.openstack.org/debtcollector/latest/index.html
Source: https://github.com/openstack/debtcollector
Version: 1.22.0
License: Apache License 2.0

Notes:
   Removed dependency on `pbr` and manually set `__version__`


psutil
------

Website: https://github.com/giampaolo/psutil
Source: https://github.com/giampaolo/psutil
Version: 5.6.7
License: BSD 3


contextvars
-------------

Source: https://github.com/MagicStack/contextvars
Version: 2.4
License: Apache License 2.0

Notes:
  - removal of metaclass usage
  - formatting
  - use a plain old dict instead of immutables.Map
  - removal of `*` syntax
"""

# Initialize `ddtrace.vendor.datadog.base.log` logger with our custom rate limited logger
# DEV: This helps ensure if there are connection issues we do not spam their logs
# DEV: Overwrite `base.log` instead of `get_logger('datadog.dogstatsd')` so we do
#      not conflict with any non-vendored datadog.dogstatsd logger
from ..internal.logger import get_logger
from .dogstatsd import base


base.log = get_logger('ddtrace.vendor.dogstatsd')



import typing

import packaging.version


def parse_version(version):
    # type: (str) -> typing.Tuple[int, int, int]
    """Convert a version string to a tuple of (major, minor, micro)

    Examples::

       1.2.3           -> (1, 2, 3)
       1.2             -> (1, 2, 0)
       1               -> (1, 0, 0)
       1.0.0-beta1     -> (1, 0, 0)
       2020.6.19       -> (2020, 6, 19)
       malformed       -> (0, 0, 0)
       10.5.0 extra    -> (10, 5, 0)
    """
    # If we have any spaces/extra text, grab the first part
    #   "1.0.0 beta1" -> "1.0.0"
    #   "1.0.0" -> "1.0.0"
    # DEV: Versions with spaces will get converted to LegacyVersion, we do this splitting
    # to maximize the chances of getting a Version as a parsing result
    if " " in version:
        version = version.split()[0]

    # version() will not raise an exception, if the version if malformed instead
    # we will end up with a LegacyVersion
    parsed = packaging.version.parse(version)

    # LegacyVersion.release will always be `None`
    if not parsed.release:
        return (0, 0, 0)

    # Version.release was added in 17.1
    # packaging >= 20.0 has `Version.{major,minor,micro}`, use the following
    # to support older versions of the library
    # https://github.com/pypa/packaging/blob/47d40f640fddb7c97b01315419b6a1421d2dedbb/packaging/version.py#L404-L417
    return (
        parsed.release[0] if len(parsed.release) >= 1 else 0,
        parsed.release[1] if len(parsed.release) >= 2 else 0,
        parsed.release[2] if len(parsed.release) >= 3 else 0,
    )



from contextlib import contextmanager
import os
import subprocess
import sys

import django
import pytest

from tests.utils import snapshot
from tests.webclient import Client


SERVER_PORT = 8000


@contextmanager
def daphne_client(django_asgi, additional_env=None):
    """Runs a django app hosted with a daphne webserver in a subprocess and
    returns a client which can be used to query it.

    Traces are flushed by invoking a tracer.shutdown() using a /shutdown-tracer route
    at the end of the testcase.
    """

    # Make sure to copy the environment as we need the PYTHONPATH and _DD_TRACE_WRITER_ADDITIONAL_HEADERS (for the test
    # token) propagated to the new process.
    env = os.environ.copy()
    env.update(additional_env or {})
    assert "_DD_TRACE_WRITER_ADDITIONAL_HEADERS" in env, "Client fixture needs test token in headers"
    env.update(
        {
            "DJANGO_SETTINGS_MODULE": "tests.contrib.django.django_app.settings",
        }
    )

    # ddtrace-run uses execl which replaces the process but the webserver process itself might spawn new processes.
    # Right now it doesn't but it's possible that it might in the future (ex. uwsgi).
    cmd = ["ddtrace-run", "daphne", "-p", str(SERVER_PORT), "tests.contrib.django.asgi:%s" % django_asgi]
    proc = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        close_fds=True,
        env=env,
    )

    client = Client("http://localhost:%d" % SERVER_PORT)

    # Wait for the server to start up
    client.wait()

    try:
        yield client
    finally:
        resp = client.get_ignored("/shutdown-tracer")
        assert resp.status_code == 200
        proc.terminate()


@pytest.mark.skipif(django.VERSION < (2, 0), reason="")
@snapshot(variants={"21x": (2, 0) <= django.VERSION < (2, 2), "": django.VERSION >= (2, 2)})
def test_urlpatterns_include(client):
    """
    When a view is specified using `django.urls.include`
        The view is traced
    """
    assert client.get("/include/test/").status_code == 200


@snapshot(
    variants={
        "18x": django.VERSION < (1, 9),
        "111x": (1, 9) <= django.VERSION < (1, 12),
        "21x": (1, 12) < django.VERSION < (2, 2),
        "": django.VERSION >= (2, 2),
    }
)
def test_middleware_trace_callable_view(client):
    # ensures that the internals are properly traced when using callable views
    assert client.get("/feed-view/").status_code == 200


@pytest.mark.skipif(
    sys.version_info >= (3, 10, 0),
    reason=("func_name changed with Python 3.10 which changes the resource name." "TODO: new snapshot required."),
)
@snapshot(
    variants={
        "18x": django.VERSION < (1, 9),
        "111x": (1, 9) <= django.VERSION < (1, 12),
        "21x": (1, 12) < django.VERSION < (2, 2),
        "": django.VERSION >= (2, 2),
    }
)
def test_middleware_trace_partial_based_view(client):
    # ensures that the internals are properly traced when using a function views
    assert client.get("/partial-view/").status_code == 200


@pytest.mark.django_db
def test_safe_string_encoding(client, snapshot_context):
    """test_safe_string_encoding.
    If we use @snapshot decorator in a Django snapshot test, the first test adds DB creation traces. Until the
    first request is executed, the SQlite DB isn't create and Django executes the migrations and the snapshot
    raises: Received unmatched spans: 'sqlite.query'
    """
    client.get("/safe-template/")
    with snapshot_context(
        variants={
            "18x": django.VERSION < (1, 9),
            "111x": (1, 9) <= django.VERSION < (1, 12),
            "21x": (1, 12) < django.VERSION < (2, 2),
            "": django.VERSION >= (2, 2),
        }
    ):
        assert client.get("/safe-template/").status_code == 200


@snapshot(
    variants={
        "18x": django.VERSION < (1, 9),
        "111x": (1, 9) <= django.VERSION < (1, 12),
        "21x": (1, 12) < django.VERSION < (2, 2),
        "": django.VERSION >= (2, 2),
    }
)
def test_404_exceptions(client):
    assert client.get("/404-view/").status_code == 404


@pytest.fixture()
def psycopg2_patched(transactional_db):
    from django.db import connections

    from ddtrace.contrib.psycopg.patch import patch
    from ddtrace.contrib.psycopg.patch import unpatch

    patch()

    # # force recreate connection to ensure psycopg2 patching has occurred
    del connections["postgres"]
    connections["postgres"].close()
    connections["postgres"].connect()

    yield

    unpatch()


@pytest.mark.django_db
def test_psycopg_query_default(client, snapshot_context, psycopg2_patched):
    """Execute a psycopg2 query on a Django database wrapper.

    If we use @snapshot decorator in a Django snapshot test, the first test adds DB creation traces
    """
    from django.db import connections
    from psycopg2.sql import SQL

    with snapshot_context(ignores=["meta.out.host"]):
        query = SQL("""select 'one' as x""")
        conn = connections["postgres"]
        with conn.cursor() as cur:
            cur.execute(query)
            rows = cur.fetchall()
            assert len(rows) == 1, rows
            assert rows[0][0] == "one"


@pytest.mark.skipif(django.VERSION < (3, 0, 0), reason="ASGI not supported in django<3")
@snapshot(
    variants={
        "30": (3, 0, 0) <= django.VERSION < (3, 1, 0),
        "31": (3, 1, 0) <= django.VERSION < (3, 2, 0),
        "3x": django.VERSION >= (3, 2, 0),
    },
    ignores=["meta.http.useragent"],
    token_override="tests.contrib.django.test_django_snapshots.test_asgi_200",
)
@pytest.mark.parametrize("django_asgi", ["application", "channels_application"])
def test_asgi_200(django_asgi):
    with daphne_client(django_asgi) as client:
        resp = client.get("/")
        assert resp.status_code == 200
        assert resp.content == b"Hello, test app."


@pytest.mark.skipif(django.VERSION < (3, 0, 0), reason="ASGI not supported in django<3")
@snapshot(ignores=["meta.http.useragent"])
def test_asgi_200_simple_app():
    # The path simple-asgi-app/ routes to an ASGI Application that is not traced
    # This test should generate an empty snapshot
    with daphne_client("channels_application") as client:
        resp = client.get("/simple-asgi-app/")
        assert resp.status_code == 200
        assert resp.content == b"Hello World. It's me simple asgi app"


@pytest.mark.skipif(django.VERSION < (3, 0, 0), reason="ASGI not supported in django<3")
@snapshot(ignores=["meta.http.useragent"])
def test_asgi_200_traced_simple_app():
    with daphne_client("channels_application") as client:
        resp = client.get("/traced-simple-asgi-app/")
        assert resp.status_code == 200
        assert resp.content == b"Hello World. It's me simple asgi app"


@pytest.mark.skipif(django.VERSION < (3, 0, 0), reason="ASGI not supported in django<3")
@snapshot(
    ignores=["meta.error.stack", "meta.http.useragent"],
    variants={
        "30": (3, 0, 0) <= django.VERSION < (3, 1, 0),
        "31": (3, 1, 0) <= django.VERSION < (3, 2, 0),
        "3x": django.VERSION >= (3, 2, 0),
    },
)
def test_asgi_500():
    with daphne_client("application") as client:
        resp = client.get("/error-500/")
        assert resp.status_code == 500


@pytest.mark.skipif(django.VERSION < (3, 2, 0), reason="Only want to test with latest Django")
@snapshot(
    ignores=[
        "meta.error.stack",
        "meta.http.request.headers.user-agent",
        "meta.http.useragent",
        "metrics._dd.appsec.waf.duration",
        "metrics._dd.appsec.waf.duration_ext",
    ]
)
def test_appsec_enabled():
    with daphne_client("application", additional_env={"DD_APPSEC_ENABLED": "true"}) as client:
        resp = client.get("/")
        assert resp.status_code == 200
        assert resp.content == b"Hello, test app."


@pytest.mark.skipif(django.VERSION < (3, 2, 0), reason="Only want to test with latest Django")
@snapshot(
    ignores=[
        "meta.error.stack",
        "meta.http.request.headers.user-agent",
        "meta.http.useragent",
        "metrics._dd.appsec.waf.duration",
        "metrics._dd.appsec.waf.duration_ext",
    ]
)
def test_appsec_enabled_attack():
    with daphne_client("application", additional_env={"DD_APPSEC_ENABLED": "true"}) as client:
        resp = client.get("/.git")
        assert resp.status_code == 404


@pytest.mark.skipif(django.VERSION < (3, 0, 0), reason="ASGI not supported in django<3")
@snapshot(
    ignores=["meta.http.useragent"],
    variants={
        "30": (3, 0, 0) <= django.VERSION < (3, 1, 0),
        "31": (3, 1, 0) <= django.VERSION < (3, 2, 0),
        "3x": django.VERSION >= (3, 2, 0),
    },
)
def test_templates_enabled():
    """Default behavior to compare with disabled variant"""
    with daphne_client("application") as client:
        resp = client.get("/template-view/")
        assert resp.status_code == 200
        assert resp.content == b"some content\n"


@pytest.mark.skipif(django.VERSION < (3, 0, 0), reason="ASGI not supported in django<3")
@snapshot(
    ignores=["meta.http.useragent"],
    variants={
        "30": (3, 0, 0) <= django.VERSION < (3, 1, 0),
        "31": (3, 1, 0) <= django.VERSION < (3, 2, 0),
        "3x": django.VERSION >= (3, 2, 0),
    },
)
def test_templates_disabled():
    """Template instrumentation disabled"""
    with daphne_client("application", additional_env={"DD_DJANGO_INSTRUMENT_TEMPLATES": "false"}) as client:
        resp = client.get("/template-view/")
        assert resp.status_code == 200
        assert resp.content == b"some content\n"



import base64
from datetime import datetime
import hashlib
import json
import logging
import re
from typing import Any
from typing import List
from typing import Mapping
from typing import Optional
from typing import Set
from typing import TYPE_CHECKING
import uuid

import attr
import cattr

import ddtrace
from ddtrace.internal import agent
from ddtrace.internal import runtime
from ddtrace.internal.utils.time import parse_isoformat


if TYPE_CHECKING:
    from typing import Callable
    from typing import MutableMapping
    from typing import Tuple

    ProductCallback = Callable[[Optional["ConfigMetadata"], Optional[Mapping[str, Any]]], None]


log = logging.getLogger(__name__)

TARGET_FORMAT = re.compile(r"^(datadog/\d+|employee)/([^/]+)/([^/]+)/([^/]+)$")


class RemoteConfigError(Exception):
    """
    An error occurred during the configuration update procedure.
    The error is reported to the agent.
    """


@attr.s(frozen=True)
class ConfigMetadata(object):
    """
    Configuration TUF target metadata
    """

    id = attr.ib(type=str)
    product_name = attr.ib(type=str)
    sha256_hash = attr.ib(type=Optional[str])
    tuf_version = attr.ib(type=Optional[int])


@attr.s
class Signature(object):
    keyid = attr.ib(type=str)
    sig = attr.ib(type=str)


@attr.s
class Key(object):
    keytype = attr.ib(type=str)
    keyid_hash_algorithms = attr.ib(type=List[str])
    keyval = attr.ib(type=Mapping)
    scheme = attr.ib(type=str)


@attr.s
class Role(object):
    keyids = attr.ib(type=List[str])
    threshold = attr.ib(type=int)


@attr.s
class Root(object):
    _type = attr.ib(type=str, validator=attr.validators.in_(("root",)))
    spec_version = attr.ib(type=str)
    consistent_snapshot = attr.ib(type=bool)
    expires = attr.ib(type=datetime, converter=parse_isoformat)
    keys = attr.ib(type=Mapping[str, Key])
    roles = attr.ib(type=Mapping[str, Role])
    version = attr.ib(type=int)


@attr.s
class SignedRoot(object):
    signatures = attr.ib(type=List[Signature])
    signed = attr.ib(type=Root)


@attr.s
class TargetDesc(object):
    length = attr.ib(type=int)
    hashes = attr.ib(type=Mapping[str, str])
    custom = attr.ib(type=Mapping[str, Any])


@attr.s
class Targets(object):
    _type = attr.ib(type=str, validator=attr.validators.in_(("targets",)))
    custom = attr.ib(type=Mapping[str, Any])
    expires = attr.ib(type=datetime, converter=parse_isoformat)
    spec_version = attr.ib(type=str, validator=attr.validators.in_(("1.0", "1.0.0")))
    targets = attr.ib(type=Mapping[str, TargetDesc])
    version = attr.ib(type=int)


@attr.s
class SignedTargets(object):
    signatures = attr.ib(type=List[Signature])
    signed = attr.ib(type=Targets)


@attr.s
class TargetFile(object):
    path = attr.ib(type=str)
    raw = attr.ib(type=str)


@attr.s
class AgentPayload(object):
    roots = attr.ib(type=List[SignedRoot], default=None)
    targets = attr.ib(type=SignedTargets, default=None)
    target_files = attr.ib(type=List[TargetFile], default=[])
    client_configs = attr.ib(type=Set[str], default={})


def _extract_target_file(payload, target, config):
    # type: (AgentPayload, str, ConfigMetadata) -> Optional[Mapping[str, Any]]
    candidates = [item.raw for item in payload.target_files if item.path == target]
    if len(candidates) != 1 or candidates[0] is None:
        log.debug("invalid target_files for %r", target)
        return None

    try:
        raw = base64.b64decode(candidates[0])
    except Exception:
        raise RemoteConfigError("invalid base64 target_files for {!r}".format(target))

    computed_hash = hashlib.sha256(raw).hexdigest()
    if computed_hash != config.sha256_hash:
        raise RemoteConfigError(
            "mismatch between target {!r} hashes {!r} != {!r}".format(target, computed_hash, config.sha256_hash)
        )

    try:
        return json.loads(raw)
    except Exception:
        raise RemoteConfigError("invalid JSON content for target {!r}".format(target))


def _parse_target(target, metadata):
    # type: (str, TargetDesc) -> ConfigMetadata
    m = TARGET_FORMAT.match(target)
    if m is None:
        raise RemoteConfigError("unexpected target format {!r}".format(target))
    _, product_name, config_id, _ = m.groups()
    return ConfigMetadata(
        id=config_id,
        product_name=product_name,
        sha256_hash=metadata.hashes.get("sha256"),
        tuf_version=metadata.custom.get("v"),
    )


class RemoteConfigClient(object):
    """
    The Remote Configuration client regularly checks for updates on the agent
    and dispatches configurations to registered products.
    """

    def __init__(self):
        # type: () -> None
        self.id = str(uuid.uuid4())
        self.agent_url = agent_url = agent.get_trace_url()
        self._conn = agent.get_connection(agent_url, timeout=agent.get_trace_agent_timeout())
        self._headers = {"content-type": "application/json"}
        self._client_tracer = dict(
            runtime_id=runtime.get_runtime_id(),
            language="python",
            tracer_version=ddtrace.__version__,
            service=ddtrace.config.service,
            env=ddtrace.config.env,
            app_version=ddtrace.config.version,
        )

        self.converter = cattr.Converter()

        def base64_to_struct(val, cls):
            raw = base64.b64decode(val)
            obj = json.loads(raw)
            return self.converter.structure_attrs_fromdict(obj, cls)

        self.converter.register_structure_hook(SignedRoot, base64_to_struct)
        self.converter.register_structure_hook(SignedTargets, base64_to_struct)

        self._products = dict()  # type: MutableMapping[str, ProductCallback]
        self._applied_configs = dict()  # type: Mapping[str, ConfigMetadata]
        self._last_targets_version = 0
        self._last_error = None  # type: Optional[str]
        self._backend_state = None  # type: Optional[str]

    def register_product(self, product_name, func=None):
        # type: (str, Optional[ProductCallback]) -> None
        if func is not None:
            self._products[product_name] = func
        else:
            self._products.pop(product_name, None)

    def _send_request(self, payload):
        # type: (str) -> Optional[Mapping[str, Any]]
        try:
            self._conn.request("POST", "v0.7/config", payload, self._headers)
            resp = self._conn.getresponse()
            data = resp.read()
        finally:
            self._conn.close()

        if resp.status == 404:
            # Remote configuration is not enabled or unsupported by the agent
            return None

        if resp.status < 200 or resp.status >= 300:
            log.warning("Unexpected error: HTTP error status %s, reason %s", resp.status, resp.reason)
            return None

        return json.loads(data)

    def _build_payload(self, state):
        # type: (Mapping[str, Any]) -> Mapping[str, Any]
        return dict(
            client=dict(
                id=self.id,
                products=list(self._products.keys()),
                is_tracer=True,
                client_tracer=self._client_tracer,
                state=state,
            ),
            cached_target_files=[],  # TODO
        )

    def _build_state(self):
        # type: () -> Mapping[str, Any]
        has_error = self._last_error is not None
        state = dict(
            root_version=1,
            targets_version=self._last_targets_version,
            config_states=[
                dict(id=config.id, version=config.tuf_version, product=config.product_name)
                for config in self._applied_configs.values()
            ],
            has_error=has_error,
        )
        if self._backend_state is not None:
            state["backend_client_state"] = self._backend_state
        if has_error:
            state["error"] = self._last_error
        return state

    def _process_targets(self, payload):
        # type: (AgentPayload) -> Tuple[Optional[int], Optional[str], Optional[Mapping[str, ConfigMetadata]]]
        if payload.targets is None:
            # no targets received
            return None, None, None

        signed = payload.targets.signed
        if signed.expires <= datetime.utcnow():
            signed_expiration = datetime.strftime(signed.expires, "%Y-%m-%dT%H:%M:%SZ")
            raise RemoteConfigError("targets are expired, expiration date was {}".format(signed_expiration))

        targets = dict()
        for target, metadata in signed.targets.items():
            config = _parse_target(target, metadata)
            if config is not None:
                targets[target] = config

        backend_state = signed.custom.get("opaque_backend_state")
        return signed.version, backend_state, targets

    def _process_response(self, data):
        # type: (Mapping[str, Any]) -> None
        try:
            # log.debug("response payload: %r", data)
            payload = self.converter.structure_attrs_fromdict(data, AgentPayload)
        except Exception:
            log.debug("invalid agent payload received: %r", data, exc_info=True)
            raise RemoteConfigError("invalid agent payload received")

        # TODO: Also check among cached targets
        paths = {_.path for _ in payload.target_files}
        if not set(payload.client_configs) <= paths:
            raise RemoteConfigError("Not all client configurations have target files")

        # 1. Deserialize targets
        last_targets_version, backend_state, targets = self._process_targets(payload)
        if last_targets_version is None or targets is None:
            log.debug("No targets in configuration payload")
            for cb in self._products.values():
                cb(None, None)
            return

        client_configs = {k: v for k, v in targets.items() if k in payload.client_configs}

        # 2. Remove previously applied configurations
        applied_configs = dict()
        for target, config in self._applied_configs.items():
            if target in client_configs and targets.get(target) == config:
                # The configuration has not changed.
                applied_configs[target] = config
                continue

            callback = self._products.get(config.product_name)
            if callback is None:
                continue

            try:
                callback(config, None)
            except Exception:
                log.debug("error while removing product %s config %r", config.product_name, config)
                continue

        # 3. Load new configurations
        for target, config in client_configs.items():
            log.debug("new configuration for product %s", config.product_name)
            callback = self._products.get(config.product_name)
            if callback is None:
                continue

            applied_config = self._applied_configs.get(target)
            if applied_config == config:
                continue

            config_content = _extract_target_file(payload, target, config)
            if config_content is None:
                continue

            try:
                callback(config, config_content)
            except Exception:
                log.debug("error while loading product %s config %r", config.product_name, config)
                continue
            else:
                applied_configs[target] = config

        self._last_targets_version = last_targets_version
        self._applied_configs = applied_configs
        self._backend_state = backend_state

    def request(self):
        # type: () -> None
        try:
            state = self._build_state()
            payload = json.dumps(self._build_payload(state), indent=2)
            # log.debug("request payload: %r", payload)
            response = self._send_request(payload)
            if response is None:
                return
            self._process_response(response)
        except RemoteConfigError as e:
            self._last_error = str(e)
            log.warning("remote configuration client reported an error", exc_info=True)
        except Exception:
            log.warning("Unexpected error", exc_info=True)
        else:
            self._last_error = None



"""The mysqldb integration instruments the mysqlclient library to trace MySQL queries.


Enabling
~~~~~~~~

The integration is enabled automatically when using
:ref:`ddtrace-run<ddtracerun>` or :func:`patch_all()<ddtrace.patch_all>`.

Or use :func:`patch()<ddtrace.patch>` to manually enable the integration::

    from ddtrace import patch
    patch(mysqldb=True)


Global Configuration
~~~~~~~~~~~~~~~~~~~~

.. py:data:: ddtrace.config.mysqldb["service"]

   The service name reported by default for spans.

   This option can also be set with the ``DD_MYSQLDB_SERVICE`` environment
   variable.

   Default: ``"mysql"``

.. py:data:: ddtrace.config.mysqldb["trace_fetch_methods"]

   Whether or not to trace fetch methods.

   Can also configured via the ``DD_MYSQLDB_TRACE_FETCH_METHODS`` environment variable.

   Default: ``False``

.. _mysqldb_config_trace_connect:

.. py:data:: ddtrace.config.mysqldb["trace_connect"]

   Whether or not to trace connecting.

   Can also be configured via the ``DD_MYSQLDB_TRACE_CONNECT`` environment variable.

   Note that if you are overriding the service name via the Pin on an individual cursor, that will not affect
   connect traces. The service name must also be overridden on the Pin on the MySQLdb module.

   Default: ``False``


Instance Configuration
~~~~~~~~~~~~~~~~~~~~~~

To configure the integration on an per-connection basis use the
``Pin`` API::

    # Make sure to import MySQLdb and not the 'connect' function,
    # otherwise you won't have access to the patched version
    from ddtrace import Pin
    import MySQLdb

    # This will report a span with the default settings
    conn = MySQLdb.connect(user="alice", passwd="b0b", host="localhost", port=3306, db="test")

    # Use a pin to override the service.
    Pin.override(conn, service='mysql-users')

    cursor = conn.cursor()
    cursor.execute("SELECT 6*7 AS the_answer;")


This package works for mysqlclient. Only the default full-Python integration works. The binary C connector provided by
_mysql is not supported.

Help on mysqlclient can be found on:
https://mysqlclient.readthedocs.io/

"""
from ...internal.utils.importlib import require_modules


required_modules = ["MySQLdb"]

with require_modules(required_modules) as missing_modules:
    if not missing_modules:
        from .patch import patch

        __all__ = ["patch"]



import multiprocessing
import os
import sys
import time


def f():
    import ddtrace.profiling.bootstrap

    profiler = ddtrace.profiling.bootstrap.profiler
    for _ in range(50):
        if any(map(len, profiler._recorder.events.values())):
            break
        time.sleep(0.1)
    # Manually stop the profiler: atexit hooks are not called in subprocesses launched by multiprocessing and we want to
    # be sure the profile are flushed out
    profiler.stop()


if __name__ == "__main__":
    print(os.getpid())
    multiprocessing.set_start_method(sys.argv[1])

    p = multiprocessing.Process(target=f)
    p.start()
    print(p.pid)
    p.join()



"""
The Datadog WSGI middleware traces all WSGI requests.


Usage
~~~~~

The middleware can be used manually via the following command::


    from ddtrace.contrib.wsgi import DDWSGIMiddleware

    # application is a WSGI application
    application = DDWSGIMiddleware(application)


Global Configuration
~~~~~~~~~~~~~~~~~~~~

.. py:data:: ddtrace.config.wsgi["service"]

   The service name reported for the WSGI application.

   This option can also be set with the ``DD_SERVICE`` environment
   variable.

   Default: ``"wsgi"``

.. py:data:: ddtrace.config.wsgi["distributed_tracing"]

   Configuration that allows distributed tracing to be enabled.

   Default: ``True``


:ref:`All HTTP tags <http-tagging>` are supported for this integration.

"""
from .wsgi import DDWSGIMiddleware


__all__ = [
    "DDWSGIMiddleware",
]



# -*- coding: utf-8 -*-
from flask import Flask
import pytest
from redis.exceptions import ConnectionError

from ddtrace.contrib.flask_cache import get_traced_cache
from ddtrace.contrib.flask_cache.tracers import CACHE_BACKEND
from ddtrace.ext import net
from tests.utils import TracerTestCase


class FlaskCacheWrapperTest(TracerTestCase):
    SERVICE = "test-flask-cache"

    def test_cache_get_without_arguments(self):
        # create the TracedCache instance for a Flask app
        Cache = get_traced_cache(self.tracer, service=self.SERVICE)
        app = Flask(__name__)
        cache = Cache(app, config={"CACHE_TYPE": "simple"})

        # make a wrong call
        with pytest.raises(TypeError) as ex:
            cache.get()

        # ensure that the error is not caused by our tracer
        assert "get()" in ex.value.args[0]
        assert "argument" in ex.value.args[0]
        spans = self.get_spans()
        # an error trace must be sent
        assert len(spans) == 1
        span = spans[0]
        assert span.service == self.SERVICE
        assert span.resource == "get"
        assert span.name == "flask_cache.cmd"
        assert span.span_type == "cache"
        assert span.error == 1

    def test_cache_set_without_arguments(self):
        # create the TracedCache instance for a Flask app
        Cache = get_traced_cache(self.tracer, service=self.SERVICE)
        app = Flask(__name__)
        cache = Cache(app, config={"CACHE_TYPE": "simple"})

        # make a wrong call
        with pytest.raises(TypeError) as ex:
            cache.set()

        # ensure that the error is not caused by our tracer
        assert "set()" in ex.value.args[0]
        assert "argument" in ex.value.args[0]
        spans = self.pop_spans()
        # an error trace must be sent
        assert len(spans) == 1
        span = spans[0]
        assert span.service == self.SERVICE
        assert span.resource == "set"
        assert span.name == "flask_cache.cmd"
        assert span.span_type == "cache"
        assert span.error == 1

    def test_cache_add_without_arguments(self):
        # create the TracedCache instance for a Flask app
        Cache = get_traced_cache(self.tracer, service=self.SERVICE)
        app = Flask(__name__)
        cache = Cache(app, config={"CACHE_TYPE": "simple"})

        # make a wrong call
        with pytest.raises(TypeError) as ex:
            cache.add()

        # ensure that the error is not caused by our tracer
        assert "add()" in ex.value.args[0]
        assert "argument" in ex.value.args[0]
        spans = self.pop_spans()
        # an error trace must be sent
        assert len(spans) == 1
        span = spans[0]
        assert span.service == self.SERVICE
        assert span.resource == "add"
        assert span.name == "flask_cache.cmd"
        assert span.span_type == "cache"
        assert span.error == 1

    def test_cache_delete_without_arguments(self):
        # create the TracedCache instance for a Flask app
        Cache = get_traced_cache(self.tracer, service=self.SERVICE)
        app = Flask(__name__)
        cache = Cache(app, config={"CACHE_TYPE": "simple"})

        # make a wrong call
        with pytest.raises(TypeError) as ex:
            cache.delete()

        # ensure that the error is not caused by our tracer
        assert "delete()" in ex.value.args[0]
        assert "argument" in ex.value.args[0]
        spans = self.pop_spans()
        # an error trace must be sent
        assert len(spans) == 1
        span = spans[0]
        assert span.service == self.SERVICE
        assert span.resource == "delete"
        assert span.name == "flask_cache.cmd"
        assert span.span_type == "cache"
        assert span.error == 1

    def test_cache_set_many_without_arguments(self):
        # create the TracedCache instance for a Flask app
        Cache = get_traced_cache(self.tracer, service=self.SERVICE)
        app = Flask(__name__)
        cache = Cache(app, config={"CACHE_TYPE": "simple"})

        # make a wrong call
        with pytest.raises(TypeError) as ex:
            cache.set_many()

        # ensure that the error is not caused by our tracer
        assert "set_many()" in ex.value.args[0]
        assert "argument" in ex.value.args[0]
        spans = self.pop_spans()
        # an error trace must be sent
        assert len(spans) == 1
        span = spans[0]
        assert span.service == self.SERVICE
        assert span.resource == "set_many"
        assert span.name == "flask_cache.cmd"
        assert span.span_type == "cache"
        assert span.error == 1

    def test_redis_cache_tracing_with_a_wrong_connection(self):
        # create the TracedCache instance for a Flask app
        Cache = get_traced_cache(self.tracer, service=self.SERVICE)
        app = Flask(__name__)
        config = {"CACHE_TYPE": "redis", "CACHE_REDIS_PORT": 2230, "CACHE_REDIS_HOST": "127.0.0.1"}
        cache = Cache(app, config=config)

        # use a wrong redis connection
        with pytest.raises(ConnectionError) as ex:
            cache.get(u"á_complex_operation")

        # ensure that the error is not caused by our tracer
        assert "127.0.0.1:2230. Connection refused." in ex.value.args[0]
        spans = self.pop_spans()
        # an error trace must be sent
        assert len(spans) == 1
        span = spans[0]
        assert span.service == self.SERVICE
        assert span.resource == "get"
        assert span.name == "flask_cache.cmd"
        assert span.span_type == "cache"
        assert span.get_tag(CACHE_BACKEND) == "redis"
        assert span.get_tag(net.TARGET_HOST) == "127.0.0.1"
        assert span.get_metric(net.TARGET_PORT) == 2230
        assert span.error == 1

    def test_memcached_cache_tracing_with_a_wrong_connection(self):
        # create the TracedCache instance for a Flask app
        Cache = get_traced_cache(self.tracer, service=self.SERVICE)
        app = Flask(__name__)
        config = {
            "CACHE_TYPE": "memcached",
            "CACHE_MEMCACHED_SERVERS": ["localhost:2230"],
        }
        cache = Cache(app, config=config)

        # use a wrong memcached connection
        try:
            cache.get(u"á_complex_operation")
        except Exception:
            pass

        # ensure that the error is not caused by our tracer
        spans = self.pop_spans()
        assert len(spans) == 1
        span = spans[0]
        assert span.service == self.SERVICE
        assert span.resource == "get"
        assert span.name == "flask_cache.cmd"
        assert span.span_type == "cache"
        assert span.get_tag(CACHE_BACKEND) == "memcached"
        assert span.get_tag(net.TARGET_HOST) == "localhost"
        assert span.get_metric(net.TARGET_PORT) == 2230

        # the pylibmc backend raises an exception and memcached backend does
        # not, so don't test anything about the status.



from ddtrace.filters import TraceFilter
from ddtrace.tracer import Tracer
from tests.utils import DummyWriter

from .utils import TornadoTestCase


class TestFilter(TraceFilter):
    def process_trace(self, trace):
        if trace[0].name == "drop":
            return None
        else:
            return trace


class TestTornadoSettings(TornadoTestCase):
    """
    Ensure that Tornado web application properly configures the given tracer.
    """

    def get_app(self):
        # Override with a real tracer
        self.tracer = Tracer()
        super(TestTornadoSettings, self).get_app()

    def get_settings(self):
        # update tracer settings
        return {
            "datadog_trace": {
                "default_service": "custom-tornado",
                "tags": {"env": "production", "debug": "false"},
                "enabled": False,
                "agent_hostname": "dd-agent.service.consul",
                "agent_port": 8126,
                "settings": {
                    "FILTERS": [
                        TestFilter(),
                    ],
                },
            },
        }

    def test_tracer_is_properly_configured(self):
        # the tracer must be properly configured
        assert self.tracer._tags.get("env") == "production"
        assert self.tracer._tags.get("debug") == "false"
        assert self.tracer.enabled is False
        assert self.tracer.agent_trace_url == "http://dd-agent.service.consul:8126"

        writer = DummyWriter()
        self.tracer.configure(enabled=True, writer=writer)
        with self.tracer.trace("keep"):
            pass
        spans = writer.pop()
        assert len(spans) == 1

        with self.tracer.trace("drop"):
            pass
        spans = writer.pop()
        assert len(spans) == 0


class TestTornadoSettingsEnabled(TornadoTestCase):
    def get_settings(self):
        return {
            "datadog_trace": {
                "default_service": "custom-tornado",
                "enabled": True,
            },
        }

    def test_service(self):
        """Ensure that the default service for a Tornado web application is configured."""
        response = self.fetch("/success/")
        assert 200 == response.code

        spans = self.get_spans()
        assert 1 == len(spans)

        assert "custom-tornado" == spans[0].service
        assert "tornado.request" == spans[0].name



from ddtrace import config
from ddtrace.vendor.wrapt import function_wrapper

from .. import trace_utils
from ...internal.utils.importlib import func_name
from ...pin import Pin
from .helpers import get_current_app


def wrap_function(instance, func, name=None, resource=None):
    """
    Helper function to wrap common flask.app.Flask methods.

    This helper will first ensure that a Pin is available and enabled before tracing
    """
    if not name:
        name = func_name(func)

    @function_wrapper
    def trace_func(wrapped, _instance, args, kwargs):
        pin = Pin._find(wrapped, _instance, instance, get_current_app())
        if not pin or not pin.enabled():
            return wrapped(*args, **kwargs)
        with pin.tracer.trace(name, service=trace_utils.int_service(pin, config.flask), resource=resource):
            return wrapped(*args, **kwargs)

    return trace_func(func)


def wrap_signal(app, signal, func):
    """
    Helper used to wrap signal handlers

    We will attempt to find the pin attached to the flask.app.Flask app
    """
    name = func_name(func)

    @function_wrapper
    def trace_func(wrapped, instance, args, kwargs):
        pin = Pin._find(wrapped, instance, app, get_current_app())
        if not pin or not pin.enabled():
            return wrapped(*args, **kwargs)

        with pin.tracer.trace(name, service=trace_utils.int_service(pin, config.flask)) as span:
            span.set_tag("flask.signal", signal)
            return wrapped(*args, **kwargs)

    return trace_func(func)



import logging
import os

from ddtrace.internal import periodic
from ddtrace.internal.logger import get_logger
from ddtrace.internal.remoteconfig.client import RemoteConfigClient
from ddtrace.internal.utils.time import StopWatch


log = get_logger(__name__)


DEFAULT_REMOTECONFIG_POLL_SECONDS = 2.0  # seconds


def get_poll_interval_seconds():
    # type:() -> float
    return float(os.getenv("DD_REMOTECONFIG_POLL_SECONDS", default=DEFAULT_REMOTECONFIG_POLL_SECONDS))


class RemoteConfigWorker(periodic.PeriodicService):
    def __init__(self):
        super(RemoteConfigWorker, self).__init__(interval=get_poll_interval_seconds())
        self._client = RemoteConfigClient()
        log.debug("RemoteConfigWorker created with polling interval %d", get_poll_interval_seconds())

    def periodic(self):
        # type: () -> None
        with StopWatch() as sw:
            self._client.request()

        t = sw.elapsed()
        if t >= self.interval:
            log_level = logging.WARNING
        else:
            log_level = logging.DEBUG
        log.log(log_level, "request config in %.5fs to %s", t, self._client.agent_url)



from libc.stdint cimport int64_t
from libc.stdint cimport uint16_t
from libc.stdint cimport uint32_t
from libc.stdint cimport uint64_t
from libcpp cimport bool


cdef extern from "include/ddwaf.h" namespace "_ddwaf_config":
    ctypedef struct _ddwaf_config_limits:
        uint32_t max_container_size
        uint32_t max_container_depth
        uint32_t max_string_length

    ctypedef struct _ddwaf_config_obfuscator:
        const char *key_regex
        const char *value_regex


cdef extern from "include/ddwaf.h":
    ctypedef struct ddwaf_version:
        uint16_t major
        uint16_t minor
        uint16_t patch

    void ddwaf_get_version(ddwaf_version *version)

    ctypedef enum DDWAF_OBJ_TYPE:
        DDWAF_OBJ_ARRAY
        DDWAF_OBJ_MAP
        DDWAF_OBJ_STRING
        DDWAF_OBJ_SIGNED
        DDWAF_OBJ_UNSIGNED
        DDWAF_OBJ_INVALID

    ctypedef struct ddwaf_object:
        const char* parameterName
        uint64_t parameterNameLength
        ddwaf_object* array
        const char* stringValue
        uint64_t intValue
        int64_t uintValue
        uint64_t nbEntries
        DDWAF_OBJ_TYPE type

    ctypedef struct ddwaf_ruleset_info:
        uint16_t loaded
        uint16_t failed
        ddwaf_object errors
        const char *version

    ctypedef struct ddwaf_config:
        _ddwaf_config_limits limits
        _ddwaf_config_obfuscator obfuscator

    ctypedef struct ddwaf_handle:
        pass

    ctypedef struct ddwaf_context:
        pass

    ctypedef enum DDWAF_RET_CODE:
        pass

    ctypedef struct ddwaf_result:
        DDWAF_RET_CODE action
        uint64_t total_runtime
        const char* data

    ctypedef void (*ddwaf_object_free_fn)(ddwaf_object *object);

    ddwaf_handle ddwaf_init(const ddwaf_object* rules, const ddwaf_config* config, ddwaf_ruleset_info *info);
    ddwaf_context ddwaf_context_init(const ddwaf_handle handle, ddwaf_object_free_fn obj_free);
    DDWAF_RET_CODE ddwaf_run(ddwaf_context context, ddwaf_object* data, ddwaf_result* result, uint64_t timeout);
    void ddwaf_context_destroy(ddwaf_context context);
    void ddwaf_result_free(ddwaf_result* result);
    void ddwaf_destroy(ddwaf_handle handle);
    const char* const* ddwaf_required_addresses(const ddwaf_handle handle, uint32_t* size);

    ctypedef enum DDWAF_LOG_LEVEL:
        DDWAF_LOG_TRACE

    ctypedef void (*ddwaf_log_fn)(DDWAF_LOG_LEVEL level, const char *function, const char *file, unsigned line, const char *message, uint64_t len);

    void ddwaf_set_log_cb(ddwaf_log_fn fn, int level);



import os

from ddtrace import Pin
from ddtrace import config
from ddtrace.vendor import wrapt

from ...ext import db
from ...ext import net
from ...internal.utils.formats import asbool
from ..dbapi import TracedConnection
from ..dbapi import TracedCursor
from ..trace_utils import unwrap


config._add(
    "snowflake",
    dict(
        _default_service="snowflake",
        # FIXME: consistent prefix span names with other dbapi integrations
        # The snowflake integration was introduced following a different pattern
        # than all other dbapi-compliant integrations. It sets span names to
        # `sql.query` whereas other dbapi-compliant integrations are set to
        # `<integration>.query`.
        _dbapi_span_name_prefix="sql",
        trace_fetch_methods=asbool(os.getenv("DD_SNOWFLAKE_TRACE_FETCH_METHODS", default=False)),
    ),
)


class _SFTracedCursor(TracedCursor):
    def _set_post_execute_tags(self, span):
        super(_SFTracedCursor, self)._set_post_execute_tags(span)
        span._set_str_tag("sfqid", self.__wrapped__.sfqid)


def patch():
    import snowflake.connector

    if getattr(snowflake.connector, "_datadog_patch", False):
        return
    setattr(snowflake.connector, "_datadog_patch", True)

    wrapt.wrap_function_wrapper(snowflake.connector, "Connect", patched_connect)
    wrapt.wrap_function_wrapper(snowflake.connector, "connect", patched_connect)


def unpatch():
    import snowflake.connector

    if getattr(snowflake.connector, "_datadog_patch", False):
        setattr(snowflake.connector, "_datadog_patch", False)

        unwrap(snowflake.connector, "Connect")
        unwrap(snowflake.connector, "connect")


def patched_connect(connect_func, _, args, kwargs):
    conn = connect_func(*args, **kwargs)
    if isinstance(conn, TracedConnection):
        return conn

    # Add default tags to each query
    tags = {
        net.TARGET_HOST: conn.host,
        net.TARGET_PORT: conn.port,
        db.NAME: conn.database,
        db.USER: conn.user,
        "db.application": conn.application,
        "db.schema": conn.schema,
        "db.warehouse": conn.warehouse,
    }

    pin = Pin(tags=tags)
    traced_conn = TracedConnection(conn, pin=pin, cfg=config.snowflake, cursor_cls=_SFTracedCursor)
    pin.onto(traced_conn)
    return traced_conn



import contextlib
import json

import pytest
import responses
import snowflake.connector

from ddtrace import Pin
from ddtrace import tracer
from ddtrace.contrib.snowflake import patch
from ddtrace.contrib.snowflake import unpatch
from tests.opentracer.utils import init_tracer
from tests.utils import override_config
from tests.utils import snapshot


if snowflake.connector.VERSION >= (2, 3, 0):
    req_mock = responses.RequestsMock(target="snowflake.connector.vendored.requests.adapters.HTTPAdapter.send")
else:
    req_mock = responses.RequestsMock(target="snowflake.connector.network.HTTPAdapter.send")


SNOWFLAKE_TYPES = {
    "TEXT": {
        "type": "TEXT",
        "name": "current_version",
        "length": None,
        "precision": None,
        "scale": None,
        "nullable": False,
    }
}


def add_snowflake_response(
    url, method=responses.POST, data=None, success=True, status=200, content_type="application/json"
):
    body = {
        "code": status,
        "success": success,
    }
    if data is not None:
        body["data"] = data

    return req_mock.add(method, url, body=json.dumps(body), status=status, content_type=content_type)


def add_snowflake_query_response(rowtype, rows, total=None):
    if total is None:
        total = len(rows)
    data = {
        "queryResponseFormat": "json",
        "rowtype": [SNOWFLAKE_TYPES[t] for t in rowtype],
        "rowset": rows,
        "total": total,
        "queryId": "mocksfqid-29244a72-90a0-4309-9168-9dcf67721f6d",
    }

    add_snowflake_response(url="https://mock-account.snowflakecomputing.com:443/queries/v1/query-request", data=data)


@contextlib.contextmanager
def _client():
    patch()
    with req_mock:
        add_snowflake_response(
            "https://mock-account.snowflakecomputing.com:443/session/v1/login-request",
            data={
                "token": "mock-token",
                "masterToken": "mock-master-token",
                "id_token": "mock-id-token",
                "sessionId": "mock-session-id",
                "sessionInfo": {
                    "databaseName": "mock-db-name",
                    "schemaName": "mock-schema-name",
                    "warehouseName": "mock-warehouse-name",
                    "roleName": "mock-role-name",
                },
            },
        )
        ctx = snowflake.connector.connect(user="mock-user", password="mock-password", account="mock-account")
    try:
        yield ctx
    finally:
        ctx.close()
        unpatch()


@pytest.fixture
def client():
    with _client() as ctx:
        yield ctx


@contextlib.contextmanager
def ot_trace():
    ot = init_tracer("snowflake_svc", tracer)
    with ot.start_active_span("snowflake_op"):
        yield


@snapshot()
@req_mock.activate
def test_snowflake_fetchone(client):
    add_snowflake_query_response(
        rowtype=["TEXT"],
        rows=[("4.30.2",)],
    )
    with client.cursor() as cur:
        res = cur.execute("select current_version();")
        assert res == cur
        assert cur.fetchone() == ("4.30.2",)


@snapshot()
@req_mock.activate
def test_snowflake_settings_override(client):
    add_snowflake_query_response(
        rowtype=["TEXT"],
        rows=[("4.30.2",)],
    )
    with override_config("snowflake", dict(service="my-snowflake-svc")):
        with client.cursor() as cur:
            res = cur.execute("select current_version();")
            assert res == cur
            assert cur.fetchone() == ("4.30.2",)


@snapshot()
@req_mock.activate
def test_snowflake_analytics_with_rate(client):
    add_snowflake_query_response(
        rowtype=["TEXT"],
        rows=[("4.30.2",)],
    )
    with override_config("snowflake", dict(analytics_enabled=True, analytics_sample_rate=0.5)):
        with client.cursor() as cur:
            res = cur.execute("select current_version();")
            assert res == cur
            assert cur.fetchone() == ("4.30.2",)


@snapshot()
@req_mock.activate
def test_snowflake_analytics_without_rate(client):
    add_snowflake_query_response(
        rowtype=["TEXT"],
        rows=[("4.30.2",)],
    )
    with override_config("snowflake", dict(analytics_enabled=True)):
        with client.cursor() as cur:
            res = cur.execute("select current_version();")
            assert res == cur
            assert cur.fetchone() == ("4.30.2",)


@pytest.mark.subprocess(env=dict(DD_SNOWFLAKE_SERVICE="env-svc"), err=None)
@snapshot()
def test_snowflake_service_env():
    from tests.contrib.snowflake.test_snowflake import _client
    from tests.contrib.snowflake.test_snowflake import add_snowflake_query_response
    from tests.contrib.snowflake.test_snowflake import req_mock

    with _client() as c:
        with req_mock:
            add_snowflake_query_response(
                rowtype=["TEXT"],
                rows=[("4.30.2",)],
            )

            with c.cursor() as cur:
                res = cur.execute("select current_version();")
                assert res == cur
                assert cur.fetchone() == ("4.30.2",)


@snapshot()
@req_mock.activate
def test_snowflake_pin_override(client):
    add_snowflake_query_response(
        rowtype=["TEXT"],
        rows=[("4.30.2",)],
    )
    Pin(service="pin-sv", tags={"custom": "tag"}).onto(client)
    with client.cursor() as cur:
        res = cur.execute("select current_version();")
        assert res == cur
        assert cur.fetchone() == ("4.30.2",)


@snapshot()
@req_mock.activate
def test_snowflake_commit(client):
    add_snowflake_query_response(rowtype=[], rows=[])
    client.commit()


@snapshot()
@req_mock.activate
def test_snowflake_rollback(client):
    add_snowflake_query_response(rowtype=[], rows=[])
    client.rollback()


@snapshot()
@req_mock.activate
def test_snowflake_fetchall(client):
    add_snowflake_query_response(
        rowtype=["TEXT"],
        rows=[("4.30.2",)],
    )
    with client.cursor() as cur:
        res = cur.execute("select current_version();")
        assert res == cur
        assert cur.fetchall() == [("4.30.2",)]


@snapshot()
@req_mock.activate
def test_snowflake_fetchall_multiple_rows(client):
    add_snowflake_query_response(
        rowtype=["TEXT", "TEXT"],
        rows=[("1a", "1b"), ("2a", "2b")],
    )
    with client.cursor() as cur:
        res = cur.execute("select a, b from t;")
        assert res == cur
        assert cur.fetchall() == [
            ("1a", "1b"),
            ("2a", "2b"),
        ]


@snapshot()
@req_mock.activate
def test_snowflake_executemany_insert(client):
    add_snowflake_query_response(
        rowtype=[],
        rows=[],
        total=2,
    )
    with client.cursor() as cur:
        res = cur.executemany(
            "insert into t (a, b) values (%s, %s);",
            [
                ("1a", "1b"),
                ("2a", "2b"),
            ],
        )
        assert res == cur
        assert res.rowcount == 2


@snapshot()
@req_mock.activate
def test_snowflake_ot_fetchone(client):
    add_snowflake_query_response(
        rowtype=["TEXT"],
        rows=[("4.30.2",)],
    )
    with ot_trace():
        with client.cursor() as cur:
            res = cur.execute("select current_version();")
            assert res == cur
            assert cur.fetchone() == ("4.30.2",)


@snapshot()
@req_mock.activate
def test_snowflake_ot_fetchall(client):
    add_snowflake_query_response(
        rowtype=["TEXT"],
        rows=[("4.30.2",)],
    )
    with ot_trace():
        with client.cursor() as cur:
            res = cur.execute("select current_version();")
            assert res == cur
            assert cur.fetchall() == [("4.30.2",)]


@snapshot()
@req_mock.activate
def test_snowflake_ot_fetchall_multiple_rows(client):
    add_snowflake_query_response(
        rowtype=["TEXT", "TEXT"],
        rows=[("1a", "1b"), ("2a", "2b")],
    )
    with ot_trace():
        with client.cursor() as cur:
            res = cur.execute("select a, b from t;")
            assert res == cur
            assert cur.fetchall() == [
                ("1a", "1b"),
                ("2a", "2b"),
            ]


@snapshot()
@req_mock.activate
def test_snowflake_ot_executemany_insert(client):
    add_snowflake_query_response(
        rowtype=[],
        rows=[],
        total=2,
    )
    with ot_trace():
        with client.cursor() as cur:
            res = cur.executemany(
                "insert into t (a, b) values (%s, %s);",
                [
                    ("1a", "1b"),
                    ("2a", "2b"),
                ],
            )
            assert res == cur
            assert res.rowcount == 2
